from django import forms

from kmuhelper.modules.settings.models import Setting
from kmuhelper.modules.settings.constants import SETTINGS_FIELDSETS
from kmuhelper.modules.settings.utils import set_db_setting


class SettingsForm(forms.Form):
    fieldsets = SETTINGS_FIELDSETS

    _fieldlist = []
    for fieldset in fieldsets:
        _fieldlist += fieldset["fields"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for obj in Setting.objects.filter(id__in=self._fieldlist):
            self.fields[obj.id] = obj.get_field()

    def save_settings(self):
        for key, value in self.cleaned_data.items():
            set_db_setting(key, value)



from django.urls import path, include, re_path

from kmuhelper import startup  # noqa
from kmuhelper import views

#######################

app_name = "kmuhelper"
urlpatterns = [
    path("", views.home, name="home"),
    path("manifest.json", views.manifest, name="manifest"),
    path("login", views.login, name="login"),
    # Developing tools
    path(
        "_templatetest/<path:templatename>", views._templatetest, name="_templatetest"
    ),
    # Main modules
    path("main/", include("kmuhelper.modules.main.urls")),
    path("api/", include("kmuhelper.modules.api.urls")),
    path("app/", include("kmuhelper.modules.app.urls")),
    path("emails/", include("kmuhelper.modules.emails.urls")),
    path("stats/", include("kmuhelper.modules.stats.urls")),
    path("settings/", include("kmuhelper.modules.settings.urls")),
    path("log", include("kmuhelper.modules.log.urls")),
    # Integrations
    path(
        "integrations/woocommerce/",
        include("kmuhelper.modules.integrations.woocommerce.urls"),
    ),
    # 404 Not found
    re_path("^.*$", views.error),
]



from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.urls import reverse_lazy
from django.shortcuts import render, redirect
from django.utils.translation import gettext_lazy, gettext

from kmuhelper.decorators import (
    require_any_kmuhelper_perms,
    require_all_kmuhelper_perms,
)
from kmuhelper.modules.settings.forms import SettingsForm

_ = gettext_lazy


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_setting"])
def settings_form(request):
    if request.method == "POST":
        form = SettingsForm(request.POST)
        if form.is_valid():
            form.save_settings()
            messages.success(request, gettext("Einstellungen gespeichert!"))
            # Redirect to prevent resending the form on reload
            return redirect("kmuhelper:settings")
    else:
        form = SettingsForm()

    return render(
        request, "kmuhelper/settings/form.html", {"form": form, "has_permission": True}
    )


@login_required(login_url=reverse_lazy("admin:login"))
@require_any_kmuhelper_perms()
def build_info(request):
    return render(request, "kmuhelper/settings/build_info.html")



from django.contrib import admin

from kmuhelper.modules.api.models import ApiKey
from kmuhelper.overrides import CustomModelAdmin


@admin.register(ApiKey)
class ApiKeyAdmin(CustomModelAdmin):
    list_display = ["id", "name", "user", "read", "write", "key_preview"]
    list_filter = ["read", "write"]

    readonly_fields = ("key",)

    fieldsets = [
        ("Settings", {"fields": ("name", "user", ("read", "write"))}),
        ("Key", {"fields": ("key",), "classes": ("collapse",)}),
    ]

    ordering = ("id",)

    HIDDEN = True


#


modeladmins = [
    (ApiKey, ApiKeyAdmin),
]



from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect
from django.urls import reverse_lazy


@login_required(login_url=reverse_lazy("admin:login"))
def log_redirect(request):
    return redirect("admin:kmuhelper_adminlogentry_changelist")



from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy

from kmuhelper.decorators import (
    confirm_action,
    require_object,
    require_all_kmuhelper_perms,
)
from kmuhelper.modules.main.models import Customer, Supplier, Supply, Order

_ = gettext_lazy

# Create your views here.


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_product", "view_supplier"])
@require_object(Supplier)
@confirm_action(_("Lieferant allen Produkten ohne Lieferant zuordnen"))
def supplier_assign(request, obj):
    count = obj.assign()
    if count == 1:
        messages.success(request, _("Lieferant wurde einem neuen Produkt zugeordnet!"))
    elif count == 0:
        messages.success(request, _("Lieferant wurde keinem neuen Produkt zugeordnet!"))
    else:
        messages.success(
            request,
            _("Lieferant wurde %d neuen Produkten einem neuen Produkt zugeordnet!")
            % count,
        )
    return redirect(reverse("admin:kmuhelper_supplier_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_product", "view_supply"])
@require_object(Supply)
@confirm_action(_("Lieferung einlagern"))
def supply_add_to_stock(request, obj):
    if obj.add_to_stock():
        messages.success(request, _("Lieferung eingelagert!"))
    else:
        messages.error(request, _("Lieferung konnte nicht eingelagert werden!"))
    return redirect(reverse("admin:kmuhelper_supply_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["add_email", "view_customer", "change_customer"])
@require_object(Customer)
def create_customer_email_registered(request, obj):
    mail = obj.create_email_registered()
    messages.success(request, _("E-Mail wurde generiert!"))
    return redirect(reverse("admin:kmuhelper_email_change", args=[mail.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["view_customer", "add_order"])
@require_object(Customer)
def create_customer_order(request, obj):
    order = Order.objects.create(customer=obj)
    messages.success(request, _("Bestellung wurde erstellt!"))
    return redirect(reverse("admin:kmuhelper_order_change", args=[order.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["add_email", "view_order", "change_order"])
@require_object(Order)
def create_order_email_invoice(request, obj):
    mail = obj.create_email_invoice()
    messages.success(request, _("E-Mail wurde generiert!"))
    return redirect(reverse("admin:kmuhelper_email_change", args=[mail.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@permission_required(["add_email", "view_order", "change_order"])
@require_object(Order)
def create_order_email_shipped(request, obj):
    mail = obj.create_email_shipped()
    messages.success(request, _("E-Mail wurde generiert!"))
    return redirect(reverse("admin:kmuhelper_email_change", args=[mail.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["add_order", "view_order"])
@require_object(Order)
@confirm_action(_("Bestellung duplizieren"))
def duplicate_order(request, obj):
    new = obj.duplicate()
    messages.success(
        request,
        _(
            "Bestellung wurde dupliziert! HINWEIS: Dies ist die neu erstellte Bestellung!"
        ),
    )
    return redirect(reverse("admin:kmuhelper_order_change", args=[new.pk]))


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["add_supply", "view_order"])
@require_object(Order)
@confirm_action(_("Bestellung zu Lieferung kopieren"))
def copy_order_to_supply(request, obj):
    new = obj.copy_to_supply()
    messages.success(request, _("Bestellung wurde zu einer Lieferung kopiert!"))
    return redirect(reverse("admin:kmuhelper_supply_change", args=[new.pk]))


def index(request):
    return redirect(reverse("admin:app_list", args=["kmuhelper"]))



from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.urls import reverse_lazy, reverse
from django.shortcuts import render, redirect
from django.utils.translation import gettext_lazy, gettext

from kmuhelper.decorators import require_object, require_all_kmuhelper_perms
from kmuhelper.modules.integrations.paymentimport.forms import CamtUploadForm
from kmuhelper.modules.integrations.paymentimport.models import PaymentImport

_ = gettext_lazy


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["add_paymentimport"])
def upload(request):
    if request.method == "POST":
        form = CamtUploadForm(request.POST, request.FILES)
        if form.is_valid():
            return form.create_payment_import_and_redirect(request)
    else:
        form = CamtUploadForm()

    return render(
        request,
        "admin/kmuhelper/paymentimport/upload.html",
        {"form": form, "has_permission": True},
    )


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_paymentimport"])
@require_object(PaymentImport)
def process(request, obj):
    if request.method == "POST":
        obj.is_processed = True
        obj.save()
        messages.success(
            request, gettext("Zahlungsimport wurde als verarbeitet markiert!")
        )
        return redirect(reverse("admin:kmuhelper_paymentimport_changelist"))

    if obj.is_processed:
        messages.warning(
            request,
            gettext(
                "Achtung! Dieser Zahlungsimport wurde bereits als verarbeitet markiert!"
            ),
        )

    ctx = obj.get_processing_context()
    return render(
        request,
        "admin/kmuhelper/paymentimport/process.html",
        {
            "has_permission": True,
            "original": obj,
            "data": ctx,
        },
    )



# Generated by Django 4.1.6 on 2023-02-28 15:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0102_paymentreceiver_improvements"),
    ]

    operations = [
        migrations.AddField(
            model_name="paymentreceiver",
            name="invoice_display_mode",
            field=models.CharField(
                choices=[
                    ("business_orders", "Geschäftlich (Bestellungen)"),
                    ("business_services", "Geschäftlich (Dienstleistungen)"),
                    ("club", "Verein"),
                    ("private", "Privat"),
                ],
                default="business_orders",
                help_text="Je nach Modus werden andere Daten auf der Rechnung angezeigt.",
                max_length=25,
                verbose_name="Anzeigemodus für Rechnungen",
            ),
        ),
    ]



from django.contrib import admin
from kmuhelper.modules.api import admin as apiadmin
from kmuhelper.modules.app import admin as appadmin
from kmuhelper.modules.emails import admin as emailadmin
from kmuhelper.modules.integrations.paymentimport import admin as paymentimportadmin
from kmuhelper.modules.log import admin as logadmin
from kmuhelper.modules.main import admin as mainadmin
from kmuhelper.modules.settings import admin as settingsadmin

# Disable "view on site" globally (for other apps too!)

admin.site.site_url = None

# List of all models and their corresponding admins

modeladmins = []

for adminclass in [
    apiadmin,
    emailadmin,
    logadmin,
    mainadmin,
    appadmin,
    settingsadmin,
    paymentimportadmin,
]:
    modeladmins.extend(adminclass.modeladmins)



from django.urls import path

from kmuhelper.modules.emails import views

#######################

urlpatterns = [
    path("<int:object_id>/view/", views.email_view, name="email-view"),
    path(
        "attachments/<int:object_id>/view/",
        views.attachment_view,
        name="attachment-view",
    ),
    path("", views.email_index, name="email-index"),
]



# Generated by Django 4.1.6 on 2023-03-11 16:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0104_order_payment_purpose"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="setting",
            options={
                "default_permissions": ("change",),
                "verbose_name": "Einstellung",
                "verbose_name_plural": "Einstellungen",
            },
        ),
        migrations.AlterModelOptions(
            name="settinghidden",
            options={
                "default_permissions": (),
                "verbose_name": "Versteckte Einstellung",
                "verbose_name_plural": "Versteckte Einstellungen",
            },
        ),
    ]



from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.translation import gettext_lazy

from kmuhelper.modules.main.models import Order
from kmuhelper.overrides import CustomModel
from kmuhelper.utils import formatprice

_ = gettext_lazy


#############


class PaymentImport(CustomModel):
    PKFILL_WIDTH = 5

    time_imported = models.DateTimeField(
        verbose_name=_("Importiert am"),
        auto_now_add=True,
    )
    is_processed = models.BooleanField(
        verbose_name=_("Verarbeitet?"),
        default=False,
    )

    data_msgid = models.CharField(
        verbose_name=_("Nachrichtenid"),
        max_length=50,
    )
    data_creationdate = models.DateTimeField(verbose_name=_("Erstellt am"))

    # Display

    @admin.display(description=_("Anzahl Einträge"))
    def entrycount(self):
        if hasattr(self, "entries"):
            return self.entries.count()

    @admin.display(description=_("Zahlungsimport"))
    def __str__(self):
        return f"{self.time_imported} ({self.pk})"

    # Methods

    def get_processing_context(self):
        context = {
            "unknown": [],
            "notfound": [],
            "ready": [],
            "alreadypaid": [],
            "unclear": [],
        }
        for entry in self.entries.order_by("ref").all():
            oid = entry.order_id()
            data = {
                "payment": entry,
                "id": oid,
            }

            if oid is None:
                context["unknown"].append(entry)
            else:
                relatedpayments = (
                    PaymentImportEntry.objects.filter(
                        amount=entry.amount,
                        currency=entry.currency,
                        iban=entry.iban,
                        ref=entry.ref,
                    )
                    .order_by("valuedate")
                    .exclude(pk=entry.pk)
                )
                if relatedpayments.count() > 0:
                    data["relatedpayments"] = relatedpayments.all()

                try:
                    order = Order.objects.get(pk=oid)

                    data["order"] = order

                    if entry.currency == "CHF" and order.is_correct_payment(
                        entry.amount, entry.valuedate
                    ):
                        if order.is_paid:
                            context["alreadypaid"].append(data)
                        else:
                            context["ready"].append(data)
                    else:
                        if order.customer:
                            data["samecustomerorders"] = order.customer.orders.exclude(
                                id=order.id
                            ).filter(is_paid=False)
                        context["unclear"].append(data)
                except ObjectDoesNotExist:
                    context["notfound"].append(data)
        return context

    class Meta:
        verbose_name = _("Zahlungsimport")
        verbose_name_plural = _("Zahlungsimporte")

    ADMIN_ICON = "fa-solid fa-file-import"


class PaymentImportEntry(models.Model):
    parent = models.ForeignKey(
        to="PaymentImport",
        on_delete=models.CASCADE,
        related_name="entries",
    )

    ref = models.CharField(
        verbose_name=_("Referenznummer"),
        max_length=50,
        default="",
    )
    additionalref = models.CharField(
        verbose_name=_("Zusätzliche Referenz"),
        max_length=250,
        default="",
    )
    iban = models.CharField(
        verbose_name=_("IBAN"),
        max_length=22,
        default="",
    )
    name = models.CharField(
        verbose_name=_("Name"),
        max_length=250,
        default="",
    )

    valuedate = models.DateField(
        verbose_name=_("Valuta"),
        null=True,
    )

    amount = models.FloatField(
        verbose_name=_("Betrag"),
    )
    currency = models.CharField(
        verbose_name=_("Währung"),
        max_length=10,
    )

    @admin.display(description=_("Betrag"), ordering="amount")
    def betrag(self):
        return formatprice(self.amount)

    @admin.display(description=_("ID"), ordering="ref")
    def order_id(self):
        if len(self.ref) == 27 and str(self.ref)[-5:-1] == "0000":
            return str(self.ref).lstrip("0")[:-5]
        return None

    @admin.display(description=_("Eintrag"))
    def __str__(self):
        return f"{self.currency} {self.amount} - {self.order_id()} - {self.name} ({self.pk})"

    class Meta:
        verbose_name = _("Zahlungsimport-Eintrag")
        verbose_name_plural = _("Zahlungsimport-Einträge")
        default_permissions = ()

    objects = models.Manager()



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from kmuhelper.modules.integrations.woocommerce.utils import is_connected


class WooCommerceStateFilter(admin.SimpleListFilter):
    """Filter that filters views based on WooCommerce state."""

    # Human-readable title which will be displayed in the
    # right admin sidebar just above the filter options.
    title = _("WooCommerce State")

    # Parameter for the filter that will be used in the URL query.
    parameter_name = "wc_state"

    def __init__(self, request, params, model, model_admin):
        super(WooCommerceStateFilter, self).__init__(
            request, params, model, model_admin
        )

    def lookups(self, request, model_admin):
        return [
            ("all", _("All")),
            ("linked", _("Linked")),
            ("not_linked", _("Not linked")),
            ("deleted", _("Deleted")),
            ("not_deleted", _("Not deleted")),
        ]

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """

        if self.value() is None:
            # default option
            self.used_parameters[self.parameter_name] = "not_deleted"

        if self.value() == "linked":
            return queryset.exclude(woocommerceid=0)
        if self.value() == "not_linked":
            return queryset.filter(woocommerceid=0)
        if self.value() == "deleted":
            return queryset.filter(woocommerce_deleted=True)
        if self.value() == "not_deleted":
            return queryset.exclude(woocommerce_deleted=True)

        return queryset

    def choices(self, changelist):
        add_facets = changelist.add_facets
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        for i, (lookup, title) in enumerate(self.lookup_choices):
            if add_facets:
                if (count := facet_counts.get(f"{i}__c", -1)) != -1:
                    title = f"{title} ({count})"
                else:
                    title = f"{title} (-)"
            yield {
                "selected": self.value() == str(lookup),
                "query_string": changelist.get_query_string(
                    {self.parameter_name: lookup}
                ),
                "display": title,
            }

    def has_output(self):
        return is_connected() and super().has_output()



import kmuhelper.modules.main.models as models
from kmuhelper.modules.integrations.woocommerce.api import product_categories
from kmuhelper.modules.integrations.woocommerce.api._base import WC_BaseObjectAPI
from kmuhelper.modules.integrations.woocommerce.api._utils import preparestring


def parse_product_name(wc_obj):
    """
    This method is a workaround because the WooCommerce REST API sometimes includes the variation options
    in the product name of the variation and sometimes doesn't:

    The attributes are not included in the title if the product has 3+ attributes or 2+ attributes and one of the
    attributes contains 2+ words.

    See source code for details:
    https://github.com/woocommerce/woocommerce/blob/43d9f10200e3d2132b472aa46b63398f79ce06ac/plugins/woocommerce/includes/data-stores/class-wc-product-variation-data-store-cpt.php#L294C1-L298C38

    To potentially solve this problem in WooCommerce directly, see
    https://wordpress.org/support/topic/help-with-variation-display-in-cart/#post-9158977
    """

    name = preparestring(wc_obj["name"])
    if wc_obj["type"] != "variation" or len(wc_obj["attributes"]) <= 1:
        return name

    attr_str = ", ".join([a["option"] for a in wc_obj["attributes"]])

    if attr_str not in name:
        return f"{name} - {attr_str}"
    return name


class WCProductsAPI(WC_BaseObjectAPI):
    LOG_PREFIX = "[deep_pink4][KMUHelper WooCommerce Products][/] -"

    MODEL = models.Product
    WC_OBJ_DOES_NOT_EXIST_CODE = "woocommerce_rest_product_invalid_id"
    WC_API_BASEURL = "products"

    def update_object_from_data(
        self, db_obj: models.Product, wc_obj: dict, is_create_event: bool = False
    ):
        db_obj.woocommerce_deleted = False

        try:
            db_obj.selling_price = float(wc_obj["regular_price"])
        except ValueError:
            self.log(
                f"[red]Failed to convert regular_price ({wc_obj['regular_price']}) to float![/]"
            )
            db_obj.selling_price = 0

        db_obj.article_number = wc_obj["sku"]
        db_obj.name = parse_product_name(wc_obj)
        db_obj.short_description = preparestring(wc_obj["short_description"])
        db_obj.description = preparestring(wc_obj["description"])
        db_obj.image_url = (
            (wc_obj["images"][0]["src"]) if len(wc_obj["images"]) > 0 else ""
        )

        if wc_obj["date_on_sale_from_gmt"]:
            db_obj.sale_from = wc_obj["date_on_sale_from_gmt"] + "+00:00"
        else:
            db_obj.sale_from = None
        if wc_obj["date_on_sale_to_gmt"]:
            db_obj.sale_to = wc_obj["date_on_sale_to_gmt"] + "+00:00"
        else:
            db_obj.sale_to = None
        if wc_obj["sale_price"]:
            db_obj.sale_price = float(wc_obj["sale_price"])
        else:
            db_obj.sale_price = None

        # Update dependencies
        self._update_dependencies(
            db_obj, wc_obj, force_update_variations=is_create_event
        )

        db_obj.save()
        self.log("Product updated:", str(db_obj))

    def create_object_from_data(self, wc_obj: dict):
        try:
            selling_price = float(wc_obj["price"])
        except ValueError:
            selling_price = 0.0

        db_obj = models.Product.objects.create(
            woocommerceid=wc_obj["id"],
            article_number=wc_obj["sku"],
            name=parse_product_name(wc_obj),
            short_description=preparestring(wc_obj["short_description"]),
            description=preparestring(wc_obj["description"]),
            selling_price=selling_price,
            image_url=(wc_obj["images"][0]["src"]) if len(wc_obj["images"]) > 0 else "",
            sale_from=(
                wc_obj["date_on_sale_from_gmt"] + "+00:00"
                if wc_obj["date_on_sale_from"]
                else None
            ),
            sale_to=(
                wc_obj["date_on_sale_to_gmt"] + "+00:00"
                if wc_obj["date_on_sale_to"]
                else None
            ),
            sale_price=(float(wc_obj["sale_price"]) if wc_obj["sale_price"] else None),
        )
        if wc_obj["manage_stock"]:
            db_obj.stock_current = wc_obj["stock_quantity"]

        # Update dependencies
        self._update_dependencies(db_obj, wc_obj)

        db_obj.save()
        self.log("Product created:", str(db_obj))
        return db_obj

    def _update_dependencies(
        self,
        db_obj: models.Product,
        wc_obj: dict,
        force_update_variations: bool = False,
    ):
        # Save before updating dependencies
        db_obj.save()

        # If product is a variation of a parent product
        if wc_obj["type"] == "variation":
            db_parent, created = models.Product.objects.get_or_create(
                woocommerceid=wc_obj["parent_id"]
            )
            if created:
                self.log("Updating parent...")
                self.update_object_from_api(db_parent)
            db_obj.parent = db_parent

        # If product is variable (has variations), also update variations
        if wc_obj["type"] == "variable" and "variations" in wc_obj:
            # Update variations
            for variation_id in wc_obj["variations"]:
                db_variation, created = models.Product.objects.get_or_create(
                    woocommerceid=variation_id
                )
                if (
                    created
                    or db_variation.woocommerce_deleted
                    or force_update_variations
                ):
                    self.log("Updating variation...")
                    self.update_object_from_api(db_variation)
            # Delete orphan variations
            for db_orphan in db_obj.children.exclude(
                woocommerceid__in=wc_obj["variations"]
            ).exclude(woocommerce_deleted=True):
                self.log("Marked orphan variation as deleted:", str(db_orphan))
                db_orphan.woocommerce_deleted = True
                db_orphan.save()

        # Categories
        db_obj.categories.clear()
        for wc_category in wc_obj["categories"]:
            db_category, created = models.ProductCategory.objects.get_or_create(
                woocommerceid=wc_category["id"]
            )
            if created:
                self.log("Created linked category! Updating...")
                product_categories.WCProductCategoriesAPI(
                    self.wcapi
                ).update_object_from_api(db_category)
            db_obj.categories.add(db_category, through_defaults=None)

    def delete_object_from_data(self, db_obj, wc_obj: dict):
        """Mark an existing object as deleted

        Returns: nothing"""

        db_obj.woocommerce_deleted = True
        db_obj.save()

        # If product has variations that have not been marked as deleted yet
        for db_child in db_obj.children.exclude(woocommerce_deleted=True):
            self.log("Marked undeleted variation as deleted:", str(db_child))
            db_child.woocommerce_deleted = True
            db_child.save()



"""Base class for PDFCreators"""

from io import BytesIO

from reportlab.platypus import SimpleDocTemplate
from reportlab.lib.units import mm
from reportlab.lib.pagesizes import A4

from django.http import FileResponse
from django.utils import translation

_ = translation.gettext


class PDFGenerator:
    def __init__(self, *args, **kwargs):
        self.elements = []

    def _build_pdf(self):
        self.__buffer = BytesIO()
        self.__doc = SimpleDocTemplate(
            self.__buffer,
            rightMargin=10 * mm,
            leftMargin=10 * mm,
            topMargin=10 * mm,
            bottomMargin=10 * mm,
            pagesize=A4,
            creator="django-kmuhelper by rafaelurben",
            author="Rafael Urben (rafaelurben.ch)",
        )

        self.__doc.build(self.elements)
        self.__buffer.seek(0)

    def get_pdf(self):
        if not hasattr(self, "__buffer"):
            self._build_pdf()
        return self.__buffer

    def get_response(self, as_attachment=False, filename="document.pdf"):
        return FileResponse(
            self.get_pdf(), as_attachment=as_attachment, filename=filename
        )



# Generated by Django 4.2.7 on 2023-11-22 21:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0105_alter_setting_options_alter_settinghidden_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="app_arrival",
            options={
                "default_permissions": [],
                "verbose_name": "Lieferung",
                "verbose_name_plural": "Lieferungen",
            },
        ),
        migrations.AlterModelOptions(
            name="app_incomingpayments",
            options={
                "default_permissions": [],
                "verbose_name": "Bestellung",
                "verbose_name_plural": "Bestellungen",
            },
        ),
        migrations.AlterModelOptions(
            name="app_shipping",
            options={
                "default_permissions": [],
                "verbose_name": "Bestellung",
                "verbose_name_plural": "Bestellungen",
            },
        ),
        migrations.AlterModelOptions(
            name="app_stock",
            options={
                "default_permissions": [],
                "verbose_name": "Produkt",
                "verbose_name_plural": "Produkte",
            },
        ),
        migrations.AlterModelOptions(
            name="app_todo",
            options={
                "default_permissions": [],
                "verbose_name": "Notiz",
                "verbose_name_plural": "Notizen",
            },
        ),
    ]



"""Utils for translations"""

from django.utils import translation

I18N_HELP_TEXT = translation.gettext_lazy(
    "Unterstützt i18n: '[:de]Deutsch[:fr]Français[:it]Italiano[:en]English[:]'"
)

# qTranslate-X compatible translation functions


def langselect(string, lang="de"):
    if "[:" + lang + "]" in string:
        return string.split("[:" + lang + "]")[1].split("[:")[0]
    if "[:de]" in string:
        return string.split("[:de]")[1].split("[:")[0]
    return string


# Autotranslations


def autotranslate_quantity_description(quantity_description, quantity=1):
    match quantity_description:
        case "Stück" | "Stück" | "Stück(e)":
            return translation.npgettext(
                "quantity description", "Stück", "Stück", quantity
            )
        case "Stunde" | "Stunden" | "Stunde(n)":
            return translation.npgettext(
                "quantity description", "Stunde", "Stunden", quantity
            )
        case "Einheit" | "Einheiten" | "Einheit(en)":
            return translation.npgettext(
                "quantity description", "Einheit", "Einheiten", quantity
            )
        case "Flasche" | "Flaschen" | "Flasche(n)":
            return translation.npgettext(
                "quantity description", "Flasche", "Flaschen", quantity
            )
        case "Tube" | "Tuben" | "Tube(n)":
            return translation.npgettext(
                "quantity description", "Tube", "Tuben", quantity
            )
    return quantity_description


def autotranslate_fee_name(name):
    match name:
        case "Versandkosten":
            return "[:de]Versandkosten[:fr]Frais d'envoi[:it]Spese di spedizione[:en]Delivery costs[:]"
    return name


# Django translation functions


class Language:
    """Context manager for language switching"""

    def __init__(self, language):
        self.language = language
        self.cur_language = None

    def __enter__(self):
        self.cur_language = translation.get_language()
        translation.activate(self.language)

    def __exit__(self, type, value, traceback):
        translation.activate(self.cur_language)



"""Useful constants for api views"""

from django.http import JsonResponse

APIKEY_INVALID = JsonResponse(
    {"error": "apikey-invalid", "message": "Your api key is invalid!"}, status=403
)

ENDPOINT_NOT_FOUND = JsonResponse(
    {"error": "endpoint-not-found", "message": "This endpoint does not exist!"},
    status=404,
)

NO_PERMISSION_APIKEY = JsonResponse(
    {
        "error": "no-permission",
        "message": "You don't have permission to access this endpoint! (authentication method: api key)",
    },
    status=403,
)

NO_PERMISSION_SESSION = JsonResponse(
    {
        "error": "no-permission",
        "message": "You don't have permission to access this endpoint! (authentication method: session)",
    },
    status=403,
)

NOT_AUTHENTICATED = JsonResponse(
    {
        "error": "not-authenticated",
        "message": "You must authenticate yourself to use this endpoint!",
    },
    status=401,
)

OBJ_NOT_FOUND = JsonResponse(
    {
        "error": "object-not-found",
        "message": "The object couln't be found!",
    },
    status=400,
)

SUCCESSFULLY_CHANGED = JsonResponse(
    {"success": "change-successfull", "message": "Successfully changed object!"},
    status=200,
)



from django.db import models
from django.utils.translation import (
    gettext_lazy,
    pgettext_lazy,
)

from kmuhelper import constants

_ = gettext_lazy


class Address:
    """Model representing an address"""

    first_name: str
    last_name: str
    company: str
    address_1: str
    address_2: str
    postcode: str
    city: str
    state: str
    country: str
    phone: str
    email: str

    def __getitem__(self, item):
        return getattr(self, item)

    def is_empty(self):
        for field in [
            "first_name",
            "last_name",
            "company",
            "address_1",
            "postcode",
            "city",
        ]:
            if getattr(self, field, "") != "":
                return False
        return True


class AddressModelMixin(models.Model):
    """This model mixin provides billing and shipping address fields and utility methods"""

    # Billing address

    @property
    def addr_billing(self) -> Address:
        addr = Address()
        for field in constants.ADDR_BILLING_FIELDS:
            setattr(addr, field.replace("addr_billing_", ""), getattr(self, field))
        return addr

    addr_billing_first_name = models.CharField(
        verbose_name=pgettext_lazy("address", "Vorname"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_billing_last_name = models.CharField(
        verbose_name=pgettext_lazy("address", "Nachname"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_billing_company = models.CharField(
        verbose_name=pgettext_lazy("address", "Firma"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_billing_address_1 = models.CharField(
        verbose_name=pgettext_lazy("address", "Adresszeile 1"),
        max_length=250,
        default="",
        blank=True,
        help_text=_('Strasse und Hausnummer oder "Postfach"'),
    )
    addr_billing_address_2 = models.CharField(
        verbose_name=pgettext_lazy("address", "Adresszeile 2"),
        max_length=250,
        default="",
        blank=True,
        help_text=_("Wird in QR-Rechnung NICHT verwendet!"),
    )
    addr_billing_city = models.CharField(
        verbose_name=pgettext_lazy("address", "Ort"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_billing_state = models.CharField(
        verbose_name=pgettext_lazy("address", "Kanton"),
        max_length=50,
        default="",
        blank=True,
    )
    addr_billing_postcode = models.CharField(
        verbose_name=pgettext_lazy("address", "Postleitzahl"),
        max_length=50,
        default="",
        blank=True,
    )
    addr_billing_country = models.CharField(
        verbose_name=pgettext_lazy("address", "Land"),
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    addr_billing_email = models.EmailField(
        verbose_name=pgettext_lazy("address", "E-Mail-Adresse"),
        blank=True,
    )
    addr_billing_phone = models.CharField(
        verbose_name=pgettext_lazy("address", "Telefon"),
        max_length=50,
        default="",
        blank=True,
    )

    # Shipping address

    @property
    def addr_shipping(self) -> Address:
        addr = Address()
        for field in constants.ADDR_SHIPPING_FIELDS:
            setattr(addr, field.replace("addr_shipping_", ""), getattr(self, field))
        return addr

    addr_shipping_first_name = models.CharField(
        verbose_name=pgettext_lazy("address", "Vorname"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_last_name = models.CharField(
        verbose_name=pgettext_lazy("address", "Nachname"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_company = models.CharField(
        verbose_name=pgettext_lazy("address", "Firma"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_address_1 = models.CharField(
        verbose_name=pgettext_lazy("address", "Adresszeile 1"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_address_2 = models.CharField(
        verbose_name=pgettext_lazy("address", "Adresszeile 2"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_city = models.CharField(
        verbose_name=pgettext_lazy("address", "Ort"),
        max_length=250,
        default="",
        blank=True,
    )
    addr_shipping_state = models.CharField(
        verbose_name=pgettext_lazy("address", "Kanton"),
        max_length=50,
        default="",
        blank=True,
    )
    addr_shipping_postcode = models.CharField(
        verbose_name=pgettext_lazy("address", "Postleitzahl"),
        max_length=50,
        default="",
        blank=True,
    )
    addr_shipping_country = models.CharField(
        verbose_name=pgettext_lazy("address", "Land"),
        max_length=2,
        default="CH",
        choices=constants.COUNTRIES,
    )
    addr_shipping_email = models.EmailField(
        verbose_name=pgettext_lazy("address", "E-Mail-Adresse"),
        blank=True,
    )
    addr_shipping_phone = models.CharField(
        verbose_name=pgettext_lazy("address", "Telefon"),
        max_length=50,
        default="",
        blank=True,
    )

    # Check function

    def addresses_are_equal(self):
        for field in constants.ADDR_FIELDS:
            if getattr(self, f"addr_billing_{field}") != getattr(
                self, f"addr_shipping_{field}"
            ):
                return False
        return True

    # Meta

    class Meta:
        abstract = True



# Generated by Django 5.0 on 2023-12-19 13:57

from django.db import migrations, models

# Context of this migration:
# Swiss VAT rates change on 2024/01/01.
# More information: https://www.estv.admin.ch/estv/en/home/value-added-tax/vat-rates-switzerland.html


MIGRATED_VAT_RATES = ((2.5, 2.6), (3.7, 3.8), (7.7, 8.1))


def migrate_vat_rates(apps, schema_editor, reverse=False):
    db_alias = schema_editor.connection.alias
    Fee = apps.get_model("kmuhelper", "Fee")
    Product = apps.get_model("kmuhelper", "Product")

    for before, after in MIGRATED_VAT_RATES:
        if reverse:
            before, after = after, before
        Fee.objects.using(db_alias).filter(vat_rate=before).update(vat_rate=after)
        Product.objects.using(db_alias).filter(vat_rate=before).update(vat_rate=after)


def migrate_vat_rates_forward(apps, schema_editor):
    migrate_vat_rates(apps, schema_editor, reverse=False)


def migrate_vat_rates_reverse(apps, schema_editor):
    migrate_vat_rates(apps, schema_editor, reverse=True)


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0107_alter_emailtemplate_mail_subject_and_more"),
    ]

    operations = [
        migrations.RunPython(
            migrate_vat_rates_forward, migrate_vat_rates_reverse, elidable=True
        ),
        migrations.AlterField(
            model_name="fee",
            name="vat_rate",
            field=models.FloatField(
                choices=[
                    (0.0, "0.0% (Mehrwertsteuerfrei)"),
                    (2.6, "2.6% (Reduzierter Satz)"),
                    (3.8, "3.8% (Sondersatz für Beherbergung)"),
                    (8.1, "8.1% (Normalsatz)"),
                    (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                    (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                    (7.7, "7.7% (Bis 2023: Normalsatz)"),
                ],
                default=8.1,
                verbose_name="MwSt-Satz",
            ),
        ),
        migrations.AlterField(
            model_name="orderfee",
            name="vat_rate",
            field=models.FloatField(
                choices=[
                    (0.0, "0.0% (Mehrwertsteuerfrei)"),
                    (2.6, "2.6% (Reduzierter Satz)"),
                    (3.8, "3.8% (Sondersatz für Beherbergung)"),
                    (8.1, "8.1% (Normalsatz)"),
                    (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                    (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                    (7.7, "7.7% (Bis 2023: Normalsatz)"),
                ],
                default=8.1,
                verbose_name="MwSt-Satz",
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="vat_rate",
            field=models.FloatField(
                choices=[
                    (0.0, "0.0% (Mehrwertsteuerfrei)"),
                    (2.6, "2.6% (Reduzierter Satz)"),
                    (3.8, "3.8% (Sondersatz für Beherbergung)"),
                    (8.1, "8.1% (Normalsatz)"),
                    (2.5, "2.5% (Bis 2023: Reduzierter Satz)"),
                    (3.7, "3.7% (Bis 2023: Sondersatz für Beherbergung)"),
                    (7.7, "7.7% (Bis 2023: Normalsatz)"),
                ],
                default=8.1,
                verbose_name="MwSt-Satz",
            ),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy

from kmuhelper.modules.main.models import Note, Order, Product, Supply

_ = gettext_lazy

#####


class App_ToDoManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(done=False)


class App_ToDo(Note):
    IS_APP_MODEL = True

    objects = App_ToDoManager()

    class Meta:
        proxy = True
        verbose_name = _("Notiz")
        verbose_name_plural = _("Notizen")
        default_permissions = []

    ADMIN_TITLE = _("ToDo-Liste")
    ADMIN_DESCRIPTION = _("Unerledigten Notizen")


class App_ShippingManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_shipped=False)


class App_Shipping(Order):
    IS_APP_MODEL = True

    objects = App_ShippingManager()

    class Meta:
        proxy = True
        verbose_name = _("Bestellung")
        verbose_name_plural = _("Bestellungen")
        default_permissions = []

    ADMIN_TITLE = _("Warenausgang")
    ADMIN_DESCRIPTION = _("Nicht versendete Bestellungen")
    ADMIN_ICON = "fa-solid fa-box-open"


class App_IncomingPaymentsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_paid=False)


class App_IncomingPayments(Order):
    IS_APP_MODEL = True

    objects = App_IncomingPaymentsManager()

    class Meta:
        proxy = True
        verbose_name = _("Bestellung")
        verbose_name_plural = _("Bestellungen")
        default_permissions = []

    ADMIN_TITLE = _("Zahlungseingang")
    ADMIN_DESCRIPTION = _("Nicht bezahlte Bestellungen")
    ADMIN_ICON = "fa-solid fa-hand-holding-dollar"


class App_Stock(Product):
    IS_APP_MODEL = True

    display_current_price = Product.display_current_price
    display_article_number = Product.display_article_number

    class Meta:
        proxy = True
        verbose_name = _("Produkt")
        verbose_name_plural = _("Produkte")
        default_permissions = []

    ADMIN_TITLE = _("Lagerbestand")
    ADMIN_DESCRIPTION = _("Alle Produkte im Lager")


class App_ArrivalManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(is_added_to_stock=False)


class App_Arrival(Supply):
    IS_APP_MODEL = True

    objects = App_ArrivalManager()

    class Meta:
        proxy = True
        verbose_name = _("Lieferung")
        verbose_name_plural = _("Lieferungen")
        default_permissions = []

    ADMIN_TITLE = _("Wareneingang")
    ADMIN_DESCRIPTION = _("Nicht eingelagerte Lieferungen")



"""
Tests admin URLs to check for unexpected 500 errors caused by mistakes in admin configs or templates.
"""

from django.conf import settings
from django.contrib.admin.models import ADDITION
from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.utils import timezone

from kmuhelper.modules.api.models import ApiKey
from kmuhelper.modules.emails.models import Attachment, EMail, EMailTemplate
from kmuhelper.modules.integrations.paymentimport.models import PaymentImport
from kmuhelper.modules.log.models import AdminLogEntry
from kmuhelper.modules.main.models import (
    ContactPerson,
    Customer,
    Fee,
    Note,
    Order,
    PaymentReceiver,
    Product,
    ProductCategory,
    Supplier,
    Supply,
)
from kmuhelper.modules.settings.models import Setting


class AdminURLTests(TestCase):
    def setUp(self):
        # Every test needs a client.
        self.client = Client()
        self.client.cookies.load({settings.LANGUAGE_COOKIE_NAME: "de"})
        self.user = User.objects.create_user(
            username="testuser", password="<PASSWORD>", is_staff=True, is_superuser=True
        )
        self.client.force_login(self.user)

    # Utils

    def _require_successful_get(self, url):
        """Checks that a GET request to a given URL is successful and returns the request object."""
        response = self.client.get(url)
        print(f"Checking GET to {url}")
        self.assertEqual(response.status_code, 200, msg=f"Couldn't open {url}")
        return response

    def _test_model_admins(
        self,
        model_name,
        pk,
        check_add=True,
        check_change=True,
        check_delete=True,
        check_history=True,
    ):
        self._require_successful_get(f"/admin/kmuhelper/{model_name}/")
        if check_add:
            self._require_successful_get(f"/admin/kmuhelper/{model_name}/add/")
        if check_change:
            self._require_successful_get(f"/admin/kmuhelper/{model_name}/{pk}/change/")
        if check_delete:
            self._require_successful_get(f"/admin/kmuhelper/{model_name}/{pk}/delete/")
        if check_history:
            self._require_successful_get(f"/admin/kmuhelper/{model_name}/{pk}/history/")

    # General tests

    def test_admin_index(self):
        response = self._require_successful_get("/admin/")
        self.assertInHTML("<h2>KMUHelper öffnen</h2>", response.content.decode())

    def test_admin_app_index(self):
        self._require_successful_get("/admin/kmuhelper/")

    # Test model admins: api

    def test_model_admin_api_key(self):
        obj = ApiKey.objects.create(name="test", user=self.user)
        self._test_model_admins("apikey", obj.pk)

    # Test model admins: emails

    def test_model_admin_attachment(self):
        obj = Attachment.objects.create()
        self._test_model_admins("attachment", obj.pk)

    def test_model_admin_email(self):
        obj = EMail.objects.create()
        self._test_model_admins("email", obj.pk)

    def test_model_admin_email_template(self):
        obj = EMailTemplate.objects.create()
        self._test_model_admins("emailtemplate", obj.pk)

    # Test model admins: log

    def test_model_admin_log_entry(self):
        obj = AdminLogEntry.objects.create(user=self.user, action_flag=ADDITION)
        self._test_model_admins(
            "adminlogentry",
            obj.pk,
            check_add=False,
            check_change=False,
            check_delete=False,
            check_history=False,
        )

    # Test model admins: main & app

    def test_model_admin_contact_person(self):
        obj = ContactPerson.objects.create(
            name="Rafael Urben", phone="+41 123 456 78 90", email="contact@tpto.ch"
        )
        self._test_model_admins("contactperson", obj.pk)

    def test_model_admin_customer(self):
        obj = Customer.objects.create()
        self._test_model_admins("customer", obj.pk)

    def test_model_admin_fee(self):
        obj = Fee.objects.create()
        self._test_model_admins("fee", obj.pk)

    def test_model_admin_note(self):
        obj = Note.objects.create()
        self._test_model_admins("note", obj.pk)
        self._test_model_admins("app_todo", obj.pk)

    def test_model_admin_order(self):
        obj = Order.objects.create(
            contact_person=ContactPerson.objects.create(
                name="X X", phone="x", email="x@x"
            ),
            payment_receiver=PaymentReceiver.objects.create(),
        )
        self._test_model_admins("order", obj.pk)
        self._test_model_admins("app_shipping", obj.pk)
        self._test_model_admins("app_incomingpayments", obj.pk)

    def test_model_admin_payment_receiver(self):
        obj = PaymentReceiver.objects.create()
        self._test_model_admins("paymentreceiver", obj.pk)

    def test_model_admin_product(self):
        obj = Product.objects.create()
        self._test_model_admins("product", obj.pk)
        self._test_model_admins("app_stock", obj.pk)

    def test_model_admin_product_category(self):
        obj = ProductCategory.objects.create()
        self._test_model_admins("productcategory", obj.pk)

    def test_model_admin_supplier(self):
        obj = Supplier.objects.create()
        self._test_model_admins("supplier", obj.pk)

    def test_model_admin_supply(self):
        obj = Supply.objects.create()
        self._test_model_admins("supply", obj.pk)
        self._test_model_admins("app_arrival", obj.pk)

    # Test model admins: payment import

    def test_model_admin_payment_import(self):
        obj = PaymentImport.objects.create(data_creationdate=timezone.now())
        self._test_model_admins("paymentimport", obj.pk, check_add=False)

    # Test model admins: settings

    def test_model_admin_setting(self):
        obj = Setting.objects.get(pk="email-signature")
        self._test_model_admins("setting", obj.pk, check_add=False, check_delete=False)



import kmuhelper.modules.main.models as models
from kmuhelper.modules.integrations.woocommerce.api._base import WC_BaseObjectAPI


class WCCustomersAPI(WC_BaseObjectAPI):
    LOG_PREFIX = "[deep_pink4][KMUHelper WooCommerce Customers][/] -"

    MODEL = models.Customer
    WC_OBJ_DOES_NOT_EXIST_CODE = "woocommerce_rest_invalid_id"
    WC_API_BASEURL = "customers"

    def update_object_from_data(self, db_obj, wc_obj: dict):
        db_obj.woocommerce_deleted = False

        db_obj.email = wc_obj["email"]
        db_obj.first_name = wc_obj["first_name"]
        db_obj.last_name = wc_obj["last_name"]
        db_obj.company = (
            wc_obj["billing"]["company"]
            if wc_obj["billing"]["company"]
            else db_obj.company
        )
        db_obj.username = wc_obj["username"]
        db_obj.avatar_url = wc_obj["avatar_url"]

        db_obj.addr_billing_first_name = wc_obj["billing"]["first_name"]
        db_obj.addr_billing_last_name = wc_obj["billing"]["last_name"]
        db_obj.addr_billing_company = wc_obj["billing"]["company"]
        db_obj.addr_billing_address_1 = wc_obj["billing"]["address_1"]
        db_obj.addr_billing_address_2 = wc_obj["billing"]["address_2"]
        db_obj.addr_billing_city = wc_obj["billing"]["city"]
        db_obj.addr_billing_state = wc_obj["billing"]["state"]
        db_obj.addr_billing_postcode = wc_obj["billing"]["postcode"]
        db_obj.addr_billing_country = wc_obj["billing"]["country"]
        db_obj.addr_billing_email = wc_obj["billing"]["email"]
        db_obj.addr_billing_phone = wc_obj["billing"]["phone"]

        db_obj.addr_shipping_first_name = wc_obj["shipping"]["first_name"]
        db_obj.addr_shipping_last_name = wc_obj["shipping"]["last_name"]
        db_obj.addr_shipping_company = wc_obj["shipping"]["company"]
        db_obj.addr_shipping_address_1 = wc_obj["shipping"]["address_1"]
        db_obj.addr_shipping_address_2 = wc_obj["shipping"]["address_2"]
        db_obj.addr_shipping_city = wc_obj["shipping"]["city"]
        db_obj.addr_shipping_state = wc_obj["shipping"]["state"]
        db_obj.addr_shipping_postcode = wc_obj["shipping"]["postcode"]
        db_obj.addr_shipping_country = wc_obj["shipping"]["country"]
        db_obj.save()

        self.log("Customer updated:", str(db_obj))

    def create_object_from_data(self, wc_obj: dict):
        db_obj = models.Customer.objects.create(
            woocommerceid=wc_obj["id"],
            email=wc_obj["email"],
            first_name=wc_obj["first_name"],
            last_name=wc_obj["last_name"],
            company=wc_obj["billing"]["company"],
            username=wc_obj["username"],
            avatar_url=wc_obj["avatar_url"],
            addr_billing_first_name=wc_obj["billing"]["first_name"],
            addr_billing_last_name=wc_obj["billing"]["last_name"],
            addr_billing_company=wc_obj["billing"]["company"],
            addr_billing_address_1=wc_obj["billing"]["address_1"],
            addr_billing_address_2=wc_obj["billing"]["address_2"],
            addr_billing_city=wc_obj["billing"]["city"],
            addr_billing_state=wc_obj["billing"]["state"],
            addr_billing_postcode=wc_obj["billing"]["postcode"],
            addr_billing_country=wc_obj["billing"]["country"],
            addr_billing_email=wc_obj["billing"]["email"],
            addr_billing_phone=wc_obj["billing"]["phone"],
            addr_shipping_first_name=wc_obj["shipping"]["first_name"],
            addr_shipping_last_name=wc_obj["shipping"]["last_name"],
            addr_shipping_company=wc_obj["shipping"]["company"],
            addr_shipping_address_1=wc_obj["shipping"]["address_1"],
            addr_shipping_address_2=wc_obj["shipping"]["address_2"],
            addr_shipping_city=wc_obj["shipping"]["city"],
            addr_shipping_state=wc_obj["shipping"]["state"],
            addr_shipping_postcode=wc_obj["shipping"]["postcode"],
            addr_shipping_country=wc_obj["shipping"]["country"],
            addr_shipping_email="",
            addr_shipping_phone="",
        )

        self.log("Customer created:", str(db_obj))
        return db_obj



import uuid

from functools import wraps

from kmuhelper.modules.api.constants import (
    NO_PERMISSION_APIKEY,
    APIKEY_INVALID,
    NO_PERMISSION_SESSION,
    NOT_AUTHENTICATED,
    OBJ_NOT_FOUND,
)
from kmuhelper.modules.api.models import ApiKey
from kmuhelper.decorators import require_object as original_require_object


def _is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


def protected(read=False, write=False, perms_required=[]):
    """Decorator: Protect an api view from unauthorized access."""

    def decorator(function):
        @wraps(function)
        def wrap(request, *args, **kwargs):
            apikey = request.GET.get("apikey", None)

            if isinstance(perms_required, str):
                perms = [perms_required]
            else:
                perms = perms_required

            perms = [f"kmuhelper.{perm}" if not "." in perm else perm for perm in perms]

            if apikey:
                if (
                    _is_valid_uuid(apikey)
                    and ApiKey.objects.filter(key=apikey).exists()
                ):
                    keyobject = ApiKey.objects.get(key=apikey)

                    if (
                        ((not read) or keyobject.read)
                        and ((not write) or keyobject.write)
                        and keyobject.has_perms(perms)
                    ):
                        return function(request, *args, **kwargs)

                    return NO_PERMISSION_APIKEY

                return APIKEY_INVALID

            if request.user.is_authenticated:
                if request.user.has_perms(perms):
                    return function(request, *args, **kwargs)

                return NO_PERMISSION_SESSION

            return NOT_AUTHENTICATED

        return wrap

    return decorator


# Shortcuts


def api_read(perms_required=[]):
    """Decorator: Requires a read api key or a logged in user to access this view"""
    return protected(read=True, write=False, perms_required=perms_required)


def api_write(perms_required=[]):
    """Decorator: Requires a write api key or a logged in user to access this view"""
    return protected(read=False, write=True, perms_required=perms_required)


def api_readwrite(perms_required=[]):
    """Decorator: Requires a read/write api key or a logged in user to access this view"""
    return protected(read=True, write=True, perms_required=perms_required)


def require_object(model):
    """Decorator: Requires an object with gived ID and passes it to the view"""
    return original_require_object(model, custom_response=OBJ_NOT_FOUND)



import abc

from django.contrib import messages
from django.utils.translation import ngettext
from kmuhelper import settings
from rich import print
from rich.progress import Progress
from woocommerce import API as WCAPI


def create_wc_api_object():
    """Create a API object from data stored in settings"""
    return WCAPI(
        url=settings.get_secret_db_setting("wc-url"),
        consumer_key=settings.get_secret_db_setting("wc-consumer_key"),
        consumer_secret=settings.get_secret_db_setting("wc-consumer_secret"),
    )


class WC_BaseAPI(abc.ABC):
    def __init__(self, wcapi: WCAPI = None):
        self.wcapi = wcapi or create_wc_api_object()

    # Logging

    LOG_PREFIX = NotImplemented

    def log(self, string, *args):
        print(self.LOG_PREFIX, string, *args)


class WC_BaseObjectAPI(WC_BaseAPI, abc.ABC):
    """Base CRUD class for WC API"""

    MODEL = NotImplemented
    WC_OBJ_DOES_NOT_EXIST_CODE = NotImplemented
    WC_API_BASEURL = NotImplemented

    # Object methods

    @abc.abstractmethod
    def create_object_from_data(self, wc_obj: dict):
        """Create a new object from WooCommerce data

        Must be implemented in subclass.

        Returns: The created database object
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def update_object_from_data(self, db_obj, wc_obj: dict):
        """Update an existing object with WooCommerce data

        Must be implemented in subclass.

        Returns: nothing"""

        raise NotImplementedError()

    def delete_object_from_data(self, db_obj, wc_obj: dict):
        """Mark an existing object as deleted

        Should be overridden by subclasses if there are dependencies that need to be removed.

        Returns: nothing"""

        db_obj.woocommerce_deleted = True
        db_obj.save()

    def update_object_from_api(self, db_object) -> bool:
        """Update a specific product from WooCommerce"""

        response = self.wcapi.get(f"{self.WC_API_BASEURL}/{db_object.woocommerceid}")
        wc_obj = response.json()

        if response.status_code != 200:
            self.log(
                "[red]Object update failed with status code {}".format(
                    response.status_code
                )
            )
            if "code" in wc_obj and wc_obj["code"] == self.WC_OBJ_DOES_NOT_EXIST_CODE:
                self.log(
                    "[red]Object does not exist in WooCommerce![/] Link removed.",
                    str(db_object),
                )
                db_object.woocommerceid = 0
                db_object.woocommerce_deleted = True
                db_object.save()
            else:
                self.log(wc_obj)
            return False

        try:
            self.update_object_from_data(db_object, wc_obj)
        except Exception as e:
            self.log("[red]Object update failed with exception {}".format(e))
            return False
        return True

    def bulk_update_objects_from_api(
        self, db_queryset, request=None
    ) -> (int, int, int):
        """Update every object in a queryset"""

        success_count = 0
        warning_count = 0
        error_count = 0

        with Progress() as progress:
            task = progress.add_task(
                self.LOG_PREFIX + " [orange_red1]Updating objects...",
                total=db_queryset.count(),
            )
            for db_obj in db_queryset:
                if db_obj.woocommerceid:
                    try:
                        self.update_object_from_api(db_obj)
                        success_count += 1
                    except Exception as e:
                        self.log(
                            "[red]Object update failed with exception {}".format(e)
                        )
                        error_count += 1
                else:
                    warning_count += 1
                progress.update(task, advance=1)
            progress.stop_task(task)

        if request is not None:
            self._add_request_messages_from_update_counts(
                request, success_count, warning_count, error_count
            )

        return success_count, warning_count, error_count

    def _post_process_imported_objects(
        self, db_obj__wc_obj_list: list[tuple[object, dict]]
    ):
        """Post-process imported objects after all objects have been imported"""
        ...

    def import_all_objects_from_api(self, request=None) -> int:
        """Import new products from WooCommerce

        Returns: number of imported products
        """

        with Progress() as progress:
            task_prepare = progress.add_task(
                self.LOG_PREFIX + " [orange_red1]Preparing object download...", total=1
            )

            excludeids = ",".join(
                [
                    str(obj.woocommerceid)
                    for obj in self.MODEL.objects.all().exclude(woocommerceid=0)
                ]
            )
            wc_objects = []

            progress.update(task_prepare, advance=1)
            progress.stop_task(task_prepare)

            task_download = progress.add_task(
                self.LOG_PREFIX + " [green]Downloading objects..."
            )

            r = self.wcapi.get(f"{self.WC_API_BASEURL}?exclude={excludeids}")
            wc_objects += r.json()

            progress.update(
                task_download, advance=1, total=int(r.headers["X-WP-TotalPages"])
            )

            for page in range(2, int(r.headers["X-WP-TotalPages"]) + 1):
                wc_objects += self.wcapi.get(
                    f"{self.WC_API_BASEURL}?exclude={excludeids}&page={page}"
                ).json()
                progress.update(task_download, advance=1)

            progress.stop_task(task_download)

            if wc_objects:
                db_wc_object_list = []

                task_process = progress.add_task(
                    self.LOG_PREFIX + " [cyan]Processing objects...",
                    total=len(wc_objects),
                )
                for wc_obj in wc_objects:
                    db_obj = self.create_object_from_data(wc_obj)
                    db_wc_object_list.append((db_obj, wc_obj))
                    progress.update(task_process, advance=1)
                progress.stop_task(task_process)

                self._post_process_imported_objects(db_wc_object_list)

        if request is not None:
            self._add_request_messages_from_import_counts(request, len(wc_objects))
        return len(wc_objects)

    # Messages

    def _add_request_messages_from_update_counts(
        self, request, success_count: int, warning_count: int, error_count: int
    ):
        """Create messages from bulk update counts"""

        if success_count:
            messages.success(
                request,
                ngettext(
                    "%(count)d %(name)s object was updated from WooCommerce.",
                    "%(count)d %(name)s objects were updated from WooCommerce",
                    success_count,
                )
                % {
                    "count": success_count,
                    "name": self.MODEL._meta.verbose_name,
                },
            )
        if warning_count:
            messages.warning(
                request,
                ngettext(
                    "%(count)d %(name)s object is not linked to a WooCommerce object.",
                    "%(count)d %(name)s objects are not linked to a WooCommerce object.",
                    warning_count,
                )
                % {
                    "count": warning_count,
                    "name": self.MODEL._meta.verbose_name,
                },
            )
        if error_count:
            messages.error(
                request,
                ngettext(
                    "%(count)d %(name)s object couldn't be updated from WooCommerce. Was it deleted?",
                    "%(count)d %(name)s objects couldn't be updated from WooCommerce. Were they deleted?",
                    error_count,
                )
                % {
                    "count": error_count,
                    "name": self.MODEL._meta.verbose_name,
                },
            )

    def _add_request_messages_from_import_counts(self, request, success_count):
        messages.success(
            request,
            ngettext(
                "%(count)d new %(name)s object has been imported from WooCommerce!",
                "%(count)d new %(name)s objects have been imported from WooCommerce!",
                success_count,
            )
            % {
                "count": success_count,
                "name": self.MODEL._meta.verbose_name,
            },
        )



"""PDF creator for invoices and delivery notes"""

from datetime import datetime

from django.utils.translation import pgettext
from reportlab.lib.colors import black
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import mm
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Table, TableStyle, Paragraph, Spacer, TopPadder, Flowable

from kmuhelper import settings
from kmuhelper.modules.pdfgeneration.base import PDFGenerator
from kmuhelper.modules.pdfgeneration.swiss_qr_invoice import QRInvoiceFlowable
from kmuhelper.translations import (
    autotranslate_quantity_description,
    autotranslate_fee_name,
    langselect,
)
from kmuhelper.utils import formatprice

style_default = ParagraphStyle("Normal", fontname="Helvetica")
style_bold = ParagraphStyle("Bold", fontname="Helvetica-Bold")

#####


class _PDFOrderPriceTable(Table):
    COLWIDTHS = [26 * mm, 80 * mm, 20 * mm, 20 * mm, 20 * mm, 20 * mm]

    @classmethod
    def from_order(cls, order, lang="de", show_payment_conditions=None):
        data = [
            (
                pgettext("Text on generated order PDF", "Art-Nr."),
                pgettext("Text on generated order PDF", "Bezeichnung"),
                pgettext("Text on generated order PDF", "Anzahl"),
                pgettext("Text on generated order PDF", "Einheit"),
                pgettext("Text on generated order PDF", "Preis"),
                pgettext("Text on generated order PDF", "Total"),
            )
        ]

        # Products

        h_products = 0

        for item in order.products.through.objects.filter(order=order):
            subtotal_without_discount = item.calc_subtotal_without_discount()
            data.append(
                (
                    item.article_number,
                    Paragraph(langselect(item.name, lang)),
                    str(item.quantity),
                    langselect(
                        autotranslate_quantity_description(
                            item.quantity_description, item.quantity
                        ),
                        lang,
                    ),
                    formatprice(item.product_price),
                    formatprice(subtotal_without_discount),
                )
            )
            h_products += 1
            if item.discount:
                data.append(
                    (
                        "",
                        "- " + pgettext("Text on generated order PDF", "Rabatt"),
                        str(item.discount),
                        "%",
                        formatprice(subtotal_without_discount),
                        formatprice(item.calc_discount()),
                    )
                )
                h_products += 1
            if item.note:
                data.append(("", Paragraph(f"- <b>{item.note}</b>"), "", "", "", ""))
                h_products += 1

        # Costs

        h_costs = 0

        for item in order.fees.through.objects.filter(order=order):
            data.append(
                (
                    "",
                    Paragraph(langselect(autotranslate_fee_name(item.name), lang)),
                    "",
                    "",
                    "",
                    formatprice(item.price),
                )
            )
            h_costs += 1
            if item.discount:
                data.append(
                    (
                        "",
                        "- " + pgettext("Text on generated order PDF", "Rabatt"),
                        str(item.discount),
                        "%",
                        formatprice(item.calc_subtotal_without_discount()),
                        formatprice(item.calc_discount()),
                    )
                )
                h_costs += 1
            if item.note:
                data.append(("", Paragraph(f"- <b>{item.note}</b>"), "", "", "", ""))
                h_costs += 1

        # VAT

        h_vat = 0

        vat_dict = dict(order.get_vat_dict())
        for vat_rate in vat_dict:
            if float(vat_rate) == 0.0 and not settings.get_file_setting(
                "PRINT_ZERO_VAT", False
            ):
                # Skip zero VAT if not explicitly enabled
                continue

            data.append(
                (
                    "",
                    pgettext("Text on generated order PDF", "MwSt"),
                    vat_rate,
                    "%",
                    formatprice(float(vat_dict[vat_rate])),
                    formatprice(float(vat_dict[vat_rate]) * (float(vat_rate) / 100)),
                )
            )
            h_vat += 1

        # Total & Payment conditions

        if show_payment_conditions is None:
            show_payment_conditions = settings.get_db_setting(
                "print-payment-conditions", False
            )

        if show_payment_conditions and order.payment_conditions:
            totaltext = (
                pgettext(
                    "Text on generated order PDF",
                    "Rechnungsbetrag, zahlbar netto innert %s Tagen",
                )
                % order.get_payment_conditions_data()[-1]["days"]
            )
        else:
            totaltext = pgettext("Text on generated order PDF", "Rechnungsbetrag")

        data.append(
            (
                Paragraph(f"<b>{totaltext}</b>"),
                "",
                "",
                "CHF",
                "",
                formatprice(order.cached_sum),
            )
        )

        h_paycond = 0
        if show_payment_conditions and order.payment_conditions:
            for paycond in order.get_payment_conditions_data():
                if paycond["percent"] != 0.0:
                    data.append(
                        (
                            pgettext(
                                "Text on generated order PDF",
                                "%(days)s Tage %(percent)s%% Skonto",
                            )
                            % {"days": paycond["days"], "percent": paycond["percent"]},
                            "",
                            "",
                            "CHF",
                            "",
                            formatprice(paycond["price"]),
                        )
                    )
                    h_paycond += 1

        # Style

        style = [
            # Horizontal lines
            # Header
            ("LINEABOVE", (0, 0), (-1, 0), 1, black),
            # Header/Products divider
            ("LINEBELOW", (0, 0), (-1, 0), 1, black),
            # Products/Costs divider
            ("LINEBELOW", (0, h_products), (-1, h_products), 0.5, black),
            # Costs/VAT divider
            (
                "LINEBELOW",
                (0, h_products + h_costs),
                (-1, h_products + h_costs),
                0.5,
                black,
            ),
            # VAT/Footer divider
            (
                "LINEBELOW",
                (0, h_products + h_costs + h_vat),
                (-1, h_products + h_costs + h_vat),
                1,
                black,
            ),
            # Footer
            ("LINEBELOW", (0, -1), (-1, -1), 1, black),
            # Span for total line
            ("SPAN", (0, -1 - h_paycond), (2, -1 - h_paycond)),
            # Horizontal alignment (same for all rows)
            ("ALIGN", (-1, 0), (-1, -1), "RIGHT"),
            ("ALIGN", (-2, 0), (-2, -1), "RIGHT"),
            ("ALIGN", (-4, 0), (-4, -1), "RIGHT"),
            # Vertical alignment (same for whole table)
            ("VALIGN", (0, 0), (-1, -1), "TOP"),
            # Bold total line
            ("FONTNAME", (0, -1 - h_paycond), (-1, -1 - h_paycond), "Helvetica-Bold"),
        ]

        return cls(data, repeatRows=1, style=TableStyle(style), colWidths=cls.COLWIDTHS)


class _PDFOrderProductTable(Table):
    COLWIDTHS = [36 * mm, 110 * mm, 20 * mm, 20 * mm]

    @classmethod
    def from_order(cls, order, lang="de"):
        data = [
            (
                pgettext("Text on generated order PDF", "Art-Nr."),
                pgettext("Text on generated order PDF", "Bezeichnung"),
                pgettext("Text on generated order PDF", "Anzahl"),
                pgettext("Text on generated order PDF", "Einheit"),
            )
        ]

        productcount = 0

        # Products

        for item in order.products.through.objects.filter(order=order):
            data.append(
                (
                    item.article_number,
                    Paragraph(langselect(item.name, lang)),
                    str(item.quantity),
                    langselect(
                        autotranslate_quantity_description(
                            item.quantity_description, item.quantity
                        ),
                        lang,
                    ),
                )
            )
            if item.note:
                data.append(("", Paragraph(f"- <b>{item.note}</b>"), "", ""))

            productcount += item.quantity

        # Total

        data.append(
            (
                pgettext("Text on generated order PDF", "Anzahl Produkte"),
                "",
                str(productcount),
                "",
            )
        )

        style = TableStyle(
            [
                ("LINEABOVE", (0, 0), (-1, 0), 1, black),
                ("LINEBELOW", (0, 0), (-1, 0), 1, black),
                ("LINEABOVE", (0, -1), (-1, -1), 1, black),
                ("LINEBELOW", (0, -1), (-1, -1), 1, black),
                ("ALIGN", (1, -1), (1, -1), "CENTER"),
                ("FONTNAME", (0, -1), (-1, -1), "Helvetica-Bold"),
                ("VALIGN", (0, 0), (-1, -1), "TOP"),
            ]
        )

        return cls(data, repeatRows=1, style=style, colWidths=cls.COLWIDTHS)


class _PDFOrderHeader(Flowable):
    def __init__(self, order, title, is_delivery_note=False):
        super().__init__()
        self.order = order
        self.title = str(title)
        self.is_delivery_note = is_delivery_note

    @classmethod
    def from_order(cls, order, title, is_delivery_note=False):
        elem = cls(order, title, is_delivery_note)
        elem.width = 210
        elem.height = 75
        elem._fixedWidth = 210
        elem._fixedHeight = 75
        return elem

    def get_address_lines(self) -> list:
        data = []

        if self.is_delivery_note:
            addr = self.order.addr_shipping
        else:
            addr = self.order.addr_billing

        if addr["company"]:
            data.append(addr["company"])
        if addr["first_name"] or addr["last_name"]:
            data.append(f'{addr["first_name"]} {addr["last_name"]}'.strip())
        data.append(addr["address_1"])
        if addr["address_2"]:
            data.append(addr["address_2"])
        data.append(f'{addr["postcode"]} {addr["city"]}'.strip())

        return data

    def get_header_block1(self) -> list:
        """Get titles and data for the first header block"""

        order = self.order
        recv = order.payment_receiver
        cp = order.contact_person

        data = [
            (pgettext("Text on generated order PDF", "Kontakt"), cp.name),
            (pgettext("Text on generated order PDF", "Tel."), cp.phone),
            (pgettext("Text on generated order PDF", "E-Mail"), cp.email),
        ]

        if recv.website:
            data.append((pgettext("Text on generated order PDF", "Web"), recv.website))
        if recv.swiss_uid:
            data.append(
                (pgettext("Text on generated order PDF", "MwSt."), recv.swiss_uid)
            )

        if len(data) > 4:
            # Only return last 4 elements if 4 lines are exceeded
            return data[-4:]
        return data

    def get_header_block2(self) -> list:
        """Get titles and data for the second header block"""

        order = self.order
        recv = order.payment_receiver

        match recv.invoice_display_mode:
            case "business_orders":
                data = [
                    (
                        pgettext("Text on generated order PDF", "Ihre Kundennummer"),
                        order.customer.pkfill(9) if order.customer else "-" * 9,
                    ),
                    (
                        pgettext("Text on generated order PDF", "Bestellnummer"),
                        order.pkfill(9),
                    ),
                    (
                        pgettext("Text on generated order PDF", "Bestelldatum"),
                        order.date.strftime("%d.%m.%Y"),
                    ),
                ]

                if self.is_delivery_note:
                    data.append(
                        (
                            pgettext("Text on generated order PDF", "Generiert am"),
                            datetime.now().date().strftime("%d.%m.%Y"),
                        )
                    )
                else:
                    data.append(
                        (
                            pgettext("Text on generated order PDF", "Rechnungsdatum"),
                            order.invoice_date.strftime("%d.%m.%Y"),
                        )
                    )

                return data
            case "business_services":
                data = [
                    (
                        pgettext("Text on generated order PDF", "Ihre Kundennummer"),
                        order.customer.pkfill(9) if order.customer else "-" * 9,
                    ),
                    (
                        pgettext("Text on generated order PDF", "Rechnungsnummer"),
                        order.pkfill(9),
                    ),
                    (
                        pgettext("Text on generated order PDF", "Rechnungsdatum"),
                        order.invoice_date.strftime("%d.%m.%Y"),
                    ),
                    (
                        pgettext("Text on generated order PDF", "Generiert am"),
                        datetime.now().date().strftime("%d.%m.%Y"),
                    ),
                ]

                return data
            case "club" | "private":
                data = [
                    (
                        pgettext("Text on generated order PDF", "Identifikation"),
                        order.customer.pkfill(9) if order.customer else "-" * 9,
                    ),
                    (
                        pgettext("Text on generated order PDF", "Rechnungsnummer"),
                        order.pkfill(9),
                    ),
                    (
                        pgettext("Text on generated order PDF", "Rechnungsdatum"),
                        order.invoice_date.strftime("%d.%m.%Y"),
                    ),
                    (
                        pgettext("Text on generated order PDF", "Generiert am"),
                        datetime.now().date().strftime("%d.%m.%Y"),
                    ),
                ]

                return data

    def draw_header(self):
        order = self.order
        recv = order.payment_receiver

        c = self.canv
        c.saveState()

        # Logo
        if recv.logourl:
            try:
                image = ImageReader(
                    recv.logourl
                )  # this will raise OSError if image is not available
                c.drawImage(
                    image,
                    120 * mm,
                    67 * mm,
                    width=20 * mm,
                    height=-20 * mm,
                    mask="auto",
                    anchor="nw",
                )
            except OSError:
                print(
                    "[KMUHelper PDF generation] Loading logo image for _PDFOrderHeader failed! URL: "
                    + recv.logourl
                )

        # Payment receiver name
        c.setFont("Helvetica-Bold", 14)
        c.drawString(12 * mm, 64 * mm, recv.display_name)

        # Payment receiver address
        t = c.beginText(12 * mm, 57 * mm)
        t.setFont("Helvetica", 10)
        t.textLine(recv.display_address_1)
        t.textLine(recv.display_address_2)
        c.drawText(t)

        # Header block 1: Company / contact person info
        block1 = self.get_header_block1()

        t1t = c.beginText(12 * mm, 46 * mm)  # Title
        t1t.setFont("Helvetica", 8)
        maxlen = max([len(title) for title, _ in block1])
        t1d = c.beginText((16 + maxlen * 1.35) * mm, 46 * mm)  # Data
        t1d.setFont("Helvetica", 8)

        for title, data in block1:
            t1t.textLine(title)
            t1d.textLine(data)

        c.drawText(t1t)
        c.drawText(t1d)

        # Header block 2: Customer/Order/Invoice info
        block2 = self.get_header_block2()

        t2t = c.beginText(12 * mm, 27 * mm)
        t2t.setFont("Helvetica", 12)
        t2d = c.beginText(64 * mm, 27 * mm)
        t2d.setFont("Helvetica", 12)

        for title, data in block2:
            t2t.textLine(title)
            t2d.textLine(data)

        c.drawText(t2t)
        c.drawText(t2d)

        # Customer address block
        t = c.beginText(120 * mm, 27 * mm)
        t.setFont("Helvetica", 12)

        for line in self.get_address_lines():
            t.textLine(line)

        c.drawText(t)

        # Title and date line
        c.setFont("Helvetica-Bold", 10)
        c.drawString(12 * mm, 0 * mm, self.title)

        c.setFont("Helvetica", 10)
        if len(self.title) <= 23:
            c.drawString(
                64 * mm,
                0 * mm,
                f"{order.date.year}-{order.pkfill()}"
                + (f" (Online #{order.woocommerceid})" if order.woocommerceid else ""),
            )
        else:
            c.drawString(
                120 * mm,
                0 * mm,
                f"{order.date.year}-{order.pkfill()}"
                + (f" (Online #{order.woocommerceid})" if order.woocommerceid else ""),
            )

        c.restoreState()

    def draw(self):
        self.canv.translate(-12 * mm, -40 * mm)
        self.draw_header()


class PDFOrder(PDFGenerator):
    def __init__(
        self,
        order,
        title,
        *,
        text=None,
        lang=None,
        is_delivery_note=False,
        add_cut_lines=True,
        show_payment_conditions=None,
    ):
        super().__init__()

        order.cached_sum = order.calc_total()

        lang = lang or order.language

        # Header
        self.elements = [
            _PDFOrderHeader.from_order(
                order, title=title, is_delivery_note=is_delivery_note
            ),
            Spacer(1, 48 * mm),
        ]

        # Custom text
        if text:
            self.elements += [
                Paragraph(text.replace("\n", "\n<br />")),
                Spacer(1, 10 * mm),
            ]

        # Main body
        if is_delivery_note:
            self.elements += [_PDFOrderProductTable.from_order(order, lang=lang)]
        else:
            self.elements += [
                _PDFOrderPriceTable.from_order(
                    order, lang=lang, show_payment_conditions=show_payment_conditions
                ),
                Spacer(1, 65 * mm),
                TopPadder(
                    QRInvoiceFlowable.from_order(order, add_cut_lines=add_cut_lines)
                ),
            ]



from django.urls import path

from kmuhelper.modules.stats import views

#######################

urlpatterns = [
    path("", views.stats, name="stats"),
    path("products-price", views.stats_products_price, name="stats-products-price"),
    path("best-products", views.best_products, name="stats-best-products"),
]



# Generated by Django 4.1.6 on 2023-02-28 22:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0103_paymentreceiver_invoice_display_mode"),
    ]

    operations = [
        migrations.AddField(
            model_name="order",
            name="payment_purpose",
            field=models.CharField(
                blank=True,
                default="",
                max_length=50,
                verbose_name="Zahlungszweck",
                help_text="Wird in QR-Rechnung verwendet.",
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.shortcuts import render, redirect
from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext

import kmuhelper.modules.config as config
from kmuhelper.decorators import (
    require_kmuhelper_module_perms,
)

_ = gettext

#####


def app_redirect(request):
    return redirect(reverse("kmuhelper:app-index"))


def manifest_redirect(request):
    return redirect(reverse("kmuhelper:manifest"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_kmuhelper_module_perms("app")
def app_index(request):
    return render(
        request,
        "admin/kmuhelper/_special/app-index.html",
        config.get_module_home_context(request, "app"),
    )



import string
from datetime import datetime, timedelta
from random import randint

import requests
from django.contrib import admin, messages
from django.core.validators import RegexValidator, MinValueValidator, MaxValueValidator
from django.db import models
from django.forms import ValidationError
from django.urls import reverse
from django.utils import timezone
from django.utils.html import mark_safe, format_html
from django.utils.text import format_lazy
from django.utils.translation import (
    gettext_lazy,
    gettext,
    npgettext,
    pgettext_lazy,
)
from rich import print

from kmuhelper import settings, constants
from kmuhelper.modules.emails.models import EMail, Attachment
from kmuhelper.modules.integrations.woocommerce.mixins import WooCommerceModelMixin
from kmuhelper.modules.main.mixins import AddressModelMixin
from kmuhelper.modules.pdfgeneration import PDFOrder
from kmuhelper.overrides import CustomModel
from kmuhelper.translations import langselect, I18N_HELP_TEXT, Language
from kmuhelper.utils import runden, formatprice, modulo10rekursiv, faq

_ = gettext_lazy


def log(content, *args):
    print("[deep_pink4][KMUHelper Main][/] -", content, *args)


###################


def default_delivery_title():
    datestr = datetime.now().strftime("%d.%m.%Y")
    return gettext("Lieferung vom %(date)s") % {"date": datestr}


# noinspection DuplicatedCode
def default_payment_recipient():
    if (
        PaymentReceiver.objects.exists()
        and PaymentReceiver.objects.filter(is_default=True).exists()
    ):
        return PaymentReceiver.objects.filter(is_default=True).first().pk
    # Fallback if there isn't a default payment receiver: Use the most recently used one
    if Order.objects.exists():
        newest_order = Order.objects.order_by("-date").first()
        return newest_order.payment_receiver_id
    if PaymentReceiver.objects.exists():
        return PaymentReceiver.objects.first().pk
    return None


# noinspection DuplicatedCode
def default_contact_person():
    if (
        ContactPerson.objects.exists()
        and ContactPerson.objects.filter(is_default=True).exists()
    ):
        return ContactPerson.objects.filter(is_default=True).first().pk
    # Fallback if there isn't a default contact person: Use the most recently used one
    if Order.objects.exists():
        newest_order = Order.objects.order_by("-date").first()
        return newest_order.contact_person_id
    if ContactPerson.objects.exists():
        return ContactPerson.objects.first().pk
    return None


def default_order_key():
    return "kh-" + str(randint(10000000, 99999999))


def default_payment_conditions():
    return settings.get_db_setting("default-payment-conditions")


#############


class ContactPerson(CustomModel):
    """Model representing a contact person"""

    PKFILL_WIDTH = 3

    name = models.CharField(
        verbose_name=_("Name"),
        max_length=50,
        help_text=_("Auf Rechnung ersichtlich!"),
    )
    phone = models.CharField(
        verbose_name=_("Telefon"),
        max_length=50,
        help_text=_("Auf Rechnung ersichtlich!"),
    )
    email = models.EmailField(
        verbose_name=_("E-Mail"),
        help_text=_("Auf Rechnung ersichtlich!"),
    )

    # Default

    is_default = models.BooleanField(
        verbose_name=_("Standard?"),
        default=False,
        help_text=_(
            "Aktivieren, um diese Kontaktperson als Standard zu setzen. Die Standard-Kontaktperson wird bei "
            "der Erstellung einer neuen Bestellung sowie beim Import einer Bestellung von WooCommerce verwendet."
        ),
    )

    @admin.display(description=_("Ansprechpartner"), ordering="name")
    def __str__(self):
        return f"[{self.pk}] {self.name}"

    class Meta:
        verbose_name = _("Ansprechpartner")
        verbose_name_plural = _("Ansprechpartner")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-user-tie"


class OrderFee(CustomModel):
    """Model representing fees for an order (e.g. shipping costs)"""

    # Links to other models
    order = models.ForeignKey(
        to="Order",
        on_delete=models.CASCADE,
    )
    linked_fee = models.ForeignKey(
        to="Fee",
        verbose_name=_("Verknüpfte Kosten"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        default=None,
    )

    # Custom data printed on the invoice
    note = models.CharField(
        verbose_name=_("Bemerkung"),
        default="",
        max_length=250,
        blank=True,
        help_text=_("Wird auf die Rechnung gedruckt."),
    )

    discount = models.FloatField(
        verbose_name=_("Rabatt in %"),
        default=0.0,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ],
    )

    # Data copied from linked Fee
    name = models.CharField(
        verbose_name=_("Name"),
        max_length=500,
        default=_("Zusätzliche Kosten"),
        help_text=I18N_HELP_TEXT,
    )
    price = models.FloatField(
        verbose_name=_("Preis (exkl. MwSt)"),
        default=0.0,
    )
    vat_rate = models.FloatField(
        verbose_name=_("MwSt-Satz"),
        choices=constants.VAT_RATES,
        default=constants.VAT_RATE_DEFAULT,
    )

    # Calculated data
    def calc_subtotal(self):
        return runden(self.price * ((100 - self.discount) / 100))

    def calc_subtotal_without_discount(self):
        return runden(self.price)

    def calc_discount(self):
        return runden(self.price * (self.discount / 100)) * -1

    # Display methods
    @admin.display(description=_("Name"), ordering="name")
    def clean_name(self, lang="de"):
        return langselect(self.name, lang)

    def __str__(self):
        if self.linked_fee:
            return f"[{self.pk}] {self.clean_name()} (#{self.linked_fee.pk})"
        return f"[{self.pk}] {self.clean_name()}"

    def save(self, **kwargs):
        if self.pk is None and self.linked_fee is not None:
            # Copy data from linked fee
            self.name = self.linked_fee.name
            self.vat_rate = self.linked_fee.vat_rate
            # Don't override price if already here (required for WooCommerce import)
            self.price = self.price or self.linked_fee.price
        super().save(**kwargs)

    class Meta:
        verbose_name = _("Bestellungskosten")
        verbose_name_plural = _("Bestellungskosten")

    objects = models.Manager()

    def copyto(self, order):
        self.pk = None
        self._state.adding = True
        self.order = order
        self.save()


class OrderItem(CustomModel):
    """Model representing the connection between 'Order' and 'Product'"""

    # Links to other models
    order = models.ForeignKey(
        to="Order",
        on_delete=models.CASCADE,
    )
    linked_product = models.ForeignKey(
        to="Product",
        verbose_name=_("Verknüpftes Produkt"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        default=None,
    )

    # Custom data printed on the invoice
    note = models.CharField(
        verbose_name=_("Bemerkung"),
        default="",
        max_length=250,
        blank=True,
        help_text=_("Wird auf die Rechnung gedruckt."),
    )

    quantity = models.IntegerField(
        verbose_name=_("Menge"),
        default=1,
    )
    discount = models.FloatField(
        verbose_name=_("Rabatt in %"),
        default=0,
        validators=[
            MinValueValidator(0),
            MaxValueValidator(100),
        ],
    )

    # Data copied from linked product
    article_number = models.CharField(
        verbose_name=_("Artikelnummer"),
        max_length=25,
    )
    quantity_description = models.CharField(
        verbose_name=_("Mengenbezeichnung"),
        max_length=100,
        default="Stück",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )
    name = models.CharField(
        verbose_name=_("Name"),
        max_length=500,
        default="",
        help_text=I18N_HELP_TEXT,
    )
    product_price = models.FloatField(
        verbose_name=_("Produktpreis (exkl. MwSt)"),
        default=0.0,
    )
    vat_rate = models.FloatField(
        verbose_name=_("MwSt-Satz"),
        choices=constants.VAT_RATES,
        default=constants.VAT_RATE_DEFAULT,
    )

    # Calculated data
    def calc_subtotal(self):
        return runden(
            self.product_price * self.quantity * ((100 - self.discount) / 100)
        )

    def calc_subtotal_without_discount(self):
        return runden(self.product_price * self.quantity)

    def calc_discount(self):
        return runden(self.product_price * self.quantity * (self.discount / 100)) * -1

    # Display methods
    @admin.display(description=_("Bezeichnung"), ordering="name")
    def clean_name(self, lang="de"):
        return langselect(self.name, lang)

    @admin.display(description=_("MwSt-Satz"))
    def display_vat_rate(self):
        return formatprice(self.vat_rate)

    def __str__(self):
        if self.linked_product_id:
            return (
                f"[{self.pk}] {self.quantity}x {self.clean_name()} (Art. {self.article_number}, "
                f"#{self.linked_product_id})"
            )
        return f"[{self.pk}] {self.quantity}x {self.clean_name()} (Art. {self.article_number})"

    def save(self, **kwargs):
        if self.pk is None and self.linked_product is not None:
            # Copy data from linked product
            self.article_number = self.linked_product.article_number
            self.quantity_description = self.linked_product.quantity_description
            self.name = self.linked_product.name
            self.vat_rate = self.linked_product.vat_rate
            # Don't override price if already here (required for WooCommerce import)
            self.product_price = runden(
                self.product_price or self.linked_product.get_current_price()
            )
        super().save(**kwargs)

    class Meta:
        verbose_name = _("Bestellungsposten")
        verbose_name_plural = _("Bestellungsposten")

    objects = models.Manager()

    def copyto(self, order):
        self.pk = None
        self._state.adding = True
        self.order = order
        self.save()


class Order(CustomModel, AddressModelMixin, WooCommerceModelMixin):
    """Model representing an order"""

    NOTE_RELATION = "order"

    date = models.DateTimeField(
        verbose_name=_("Datum"),
        default=timezone.now,
    )

    invoice_date = models.DateField(
        verbose_name=_("Rechnungsdatum"),
        default=None,
        blank=True,
        null=True,
        help_text=_(
            "Wird auch als Startpunkt für die Zahlungskonditionen verwendet. Wird beim Erstellen "
            "eines PDF automatisch mit dem aktuellen Datum befüllt."
        ),
    )

    payment_conditions = models.CharField(
        verbose_name=_("Zahlungskonditionen"),
        default=default_payment_conditions,
        validators=[
            RegexValidator(
                r"^([0-9]+(\.[0-9]+)?:[0-9]+;)*0:[0-9]+$",
                _(
                    "Bitte benutze folgendes Format: 'p:d;p:d' - p = Skonto in %; d = Tage"
                ),
            )
        ],
        max_length=25,
        blank=True,
        null=True,
        help_text=format_lazy(
            "{} -> {}",
            _("Skonto und Zahlungsfrist"),
            faq("wie-funktionieren-zahlungskonditionen"),
        ),
    )

    status = models.CharField(
        verbose_name=_("Status"),
        max_length=11,
        default="processing",
        choices=constants.ORDERSTATUS,
    )
    is_shipped = models.BooleanField(
        verbose_name=_("Versendet?"),
        default=False,
        help_text=format_lazy(
            "{} -> {}",
            _("Mehr Infos"),
            faq("was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere"),
        ),
    )
    shipped_on = models.DateField(
        verbose_name=_("Versendet am"),
        default=None,
        blank=True,
        null=True,
    )
    tracking_number = models.CharField(
        verbose_name=_("Trackingnummer"),
        default="",
        blank=True,
        max_length=35,
        validators=[
            RegexValidator(
                r"^[A-Z0-9\.]{8,}$",
                _(
                    "Erlaubte Zeichen: Grossbuchstaben, Zahlen und Punkte. Länge: 8-35 Zeichen"
                ),
            )
        ],
        help_text=_("Trackingnummer ohne Leerzeichen. (optional)"),
    )

    is_removed_from_stock = models.BooleanField(
        verbose_name=_("Ausgelagert?"),
        default=False,
    )

    payment_method = models.CharField(
        verbose_name=_("Zahlungsmethode"),
        max_length=7,
        default="cod",
        choices=constants.PAYMENTMETHODS,
    )
    is_paid = models.BooleanField(
        verbose_name=_("Bezahlt?"),
        default=False,
        help_text=format_lazy(
            "{} -> {}",
            _("Mehr Infos"),
            faq("was-passiert-wenn-ich-eine-bestellung-als-bezahltversendet-markiere"),
        ),
    )
    paid_on = models.DateField(
        verbose_name=_("Bezahlt am"),
        default=None,
        blank=True,
        null=True,
    )

    payment_purpose = models.CharField(
        verbose_name=_("Zahlungszweck"),
        max_length=50,
        default="",
        blank=True,
        help_text=_("Wird in QR-Rechnung verwendet."),
    )

    customer_note = models.TextField(
        verbose_name=_("Kundennotiz"),
        default="",
        blank=True,
        help_text=_("Vom Kunden erfasste Notiz."),
    )

    order_key = models.CharField(
        verbose_name=_("Bestellungs-Schlüssel"),
        max_length=50,
        blank=True,
        default=default_order_key,
    )

    customer = models.ForeignKey(
        to="Customer",
        on_delete=models.SET_NULL,
        verbose_name=_("Kunde"),
        blank=True,
        null=True,
        related_name="orders",
    )
    payment_receiver = models.ForeignKey(
        to="PaymentReceiver",
        on_delete=models.PROTECT,
        verbose_name=_("Zahlungsempfänger"),
        default=default_payment_recipient,
    )
    contact_person = models.ForeignKey(
        to="ContactPerson",
        verbose_name=_("Ansprechpartner"),
        on_delete=models.PROTECT,
        default=default_contact_person,
    )

    # Connections

    products = models.ManyToManyField(
        to="Product",
        through="OrderItem",
        through_fields=("order", "linked_product"),
    )

    fees = models.ManyToManyField(
        to="Fee",
        through="OrderFee",
        through_fields=("order", "linked_fee"),
    )

    email_link_invoice = models.ForeignKey(
        to="EMail",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
    )
    email_link_shipped = models.ForeignKey(
        to="EMail",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
    )

    cached_sum = models.FloatField(
        verbose_name=_("Total in CHF"),
        default=0.0,
    )

    # Data for PDF generation (only editable via PDF generation form)

    pdf_title = models.CharField(
        editable=False,
        default="",
        blank=True,
        max_length=32,
    )
    pdf_text = models.TextField(
        editable=False,
        default="",
        blank=True,
    )

    # Properties

    @property
    def language(self):
        if self.customer is not None:
            return self.customer.language
        return "de"

    # Functions

    def import_customer_data(self):
        "Copy the customer data from the customer into the order"

        for field in constants.ADDR_SHIPPING_FIELDS + constants.ADDR_BILLING_FIELDS:
            setattr(self, field, getattr(self.customer, field))

    def second_save(self, **kwargs):
        "This HAS to be called after all related models have been saved."

        self.cached_sum = self.calc_total()
        if self.is_shipped and (not self.is_removed_from_stock):
            for i in self.products.through.objects.filter(order=self):
                if i.linked_product is not None:
                    i.linked_product.stock_current -= i.quantity
                    i.linked_product.save()
            self.is_removed_from_stock = True

        super().save(**kwargs)

    def save(self, **kwargs):
        if not self.pk and not self.woocommerceid and self.customer:
            self.import_customer_data()

        super().save(**kwargs)

    @admin.display(description=_("Trackinglink"), ordering="tracking_number")
    def tracking_link(self):
        return (
            f"https://www.post.ch/swisspost-tracking?formattedParcelCodes={self.tracking_number}"
            if self.tracking_number
            else None
        )

    def get_unstructured_message(self):
        "Returns the unstructured message for the QR-Invoice"

        if self.payment_purpose:
            return self.payment_purpose
        if self.payment_receiver.mode == "QRR":
            return str(self.date.strftime("%d.%m.%Y"))
        return _("Referenznummer") + ": " + str(self.id)

    def get_qr_reference_number(self):
        """Returns the formatted reference number for the QR-Invoice"""

        a = self.pkfill(22) + "0000"
        b = a + str(modulo10rekursiv(a))
        c = (
            b[0:2]
            + " "
            + b[2:7]
            + " "
            + b[7:12]
            + " "
            + b[12:17]
            + " "
            + b[17:22]
            + " "
            + b[22:27]
        )
        return c

    def get_qr_billing_information(self):
        """Returns the billing information for the QR-Invoice

        Definition 1:
        https://www.swico.ch/media/filer_public/1c/cd/1ccd7062-fc69-40f8-be3f-2a3ba9048c5f/v2_qr-bill-s1-syntax-fr.pdf
        Definition 2 (page 61...):
        https://www.six-group.com/dam/download/banking-services/interbank-clearing/de/standardization/ig-qr-bill-de.pdf
        """

        date = (self.invoice_date or self.date).strftime("%y%m%d")

        output = f"//S1/10/{self.pk}/11/{date}"

        if self.payment_receiver.swiss_uid:
            uid = self.payment_receiver.swiss_uid.split("-")[1].replace(".", "")
            output += f"/30/{uid}"

        vat_dict = self.get_vat_dict()
        var_str = ";".join(f"{rate}:{vat_dict[rate]}" for rate in vat_dict)
        output += f"/31/{date}/32/{var_str}"

        if self.payment_conditions:
            output += f"/40/{self.payment_conditions}"
        return output

    def get_payment_conditions_data(self):
        """Get the payment conditions as a list of dictionaries"""

        if not self.payment_conditions:
            return []

        data = []
        for pc in self.payment_conditions.split(";"):
            percent, days = pc.split(":")
            percent, days = float(percent), int(days)
            data.append(
                {
                    "days": days,
                    "date": (self.invoice_date or self.date) + timedelta(days=days),
                    "percent": percent,
                    "price": runden(self.cached_sum * (1 - (percent / 100))),
                }
            )
        data.sort(key=lambda x: x["date"])
        return data

    def get_vat_dict(self):
        """Get the VAT as a dictionary

        Format: { rate: total, rate2: total2 }"""
        vat_dict = {}
        for p in self.products.through.objects.filter(order=self):
            if str(p.vat_rate) in vat_dict:
                vat_dict[str(p.vat_rate)] += p.calc_subtotal()
            else:
                vat_dict[str(p.vat_rate)] = p.calc_subtotal()
        for k in self.fees.through.objects.filter(order=self):
            if str(k.vat_rate) in vat_dict:
                vat_dict[str(k.vat_rate)] += k.calc_subtotal()
            else:
                vat_dict[str(k.vat_rate)] = k.calc_subtotal()
        for s in vat_dict:
            vat_dict[s] = runden(vat_dict[s])
        return vat_dict

    def calc_total_without_vat(self):
        total = 0
        for i in self.products.through.objects.filter(order=self):
            total += i.calc_subtotal()
        for i in self.fees.through.objects.filter(order=self):
            total += i.calc_subtotal()
        return runden(total)

    def calc_total_vat(self):
        total_vat = 0
        vat_dict = self.get_vat_dict()
        for vat_rate in vat_dict:
            total_vat += runden(float(vat_dict[vat_rate] * (float(vat_rate) / 100)))
        return runden(total_vat)

    def calc_total(self):
        return runden(self.calc_total_without_vat() + self.calc_total_vat())

    @admin.display(description=_("Rechnungstotal"))
    def display_total_breakdown(self):
        return f"{formatprice(self.calc_total_without_vat())} CHF + {formatprice(self.calc_total_vat())} CHF MwSt = {formatprice(self.calc_total())} CHF"

    @admin.display(description=_("Total"), ordering="cached_sum")
    def display_cached_sum(self):
        return f"{formatprice(self.cached_sum)} CHF"

    @admin.display(description=_("Name"))
    def name(self):
        return (
            (
                f"{self.date.year}-"
                if self.date and not isinstance(self.date, str)
                else ""
            )
            + (
                f"{self.pkfill()}"
                + (f" (WC#{self.woocommerceid})" if self.woocommerceid else "")
            )
            + " - "
            + self.display_customer()
        )

    @admin.display(description=_("Info"))
    def info(self):
        return f'{self.date.strftime("%d.%m.%Y")} - ' + self.display_customer()

    @admin.display(description=_("Bezahlt nach"))
    def display_paid_after(self):
        if self.paid_on is None or self.invoice_date is None:
            return "-"

        daydiff = (self.paid_on - self.invoice_date).days

        return npgettext(
            "Paid after ...",
            _("%(count)d Tag") % {"count": daydiff},
            _("%(count)d Tagen") % {"count": daydiff},
            daydiff,
        )

    @admin.display(description=_("Konditionen"))
    def display_payment_conditions(self):
        """Get the payment conditions as a multiline string of values"""

        conditions = self.get_payment_conditions_data()
        output = ""

        for condition in conditions:
            datestr = condition["date"].strftime("%d.%m.%Y")
            price = formatprice(condition["price"])
            percent = condition["percent"]
            output += f"{price} CHF " + gettext("bis") + f" {datestr} ({percent}%)<br>"

        return mark_safe(output)

    @admin.display(description=_("Kunde"), ordering="customer")
    def display_customer(self):
        if self.customer:
            return str(self.customer)

        id0 = "0".zfill(Customer.PKFILL_WIDTH)
        text = f"[{id0}] ({_('Gast')}) "
        if self.addr_billing_first_name:
            text += self.addr_billing_first_name + " "
        if self.addr_billing_last_name:
            text += self.addr_billing_last_name + " "
        if self.addr_billing_company:
            text += self.addr_billing_company + " "
        if self.addr_billing_postcode and self.addr_billing_city:
            text += f"({self.addr_billing_postcode} {self.addr_billing_city})"

        return text

    @admin.display(
        description=format_html(
            '<abbr title="{}"><i class="fa-solid fa-truck-fast"></i></abbr>',
            _("Als versendet markiert?"),
        ),
        ordering="is_shipped",
        boolean=True,
    )
    def display_is_shipped(self):
        return self.is_shipped

    @admin.display(
        description=format_html(
            '<abbr title="{}"><i class="fa-solid fa-hand-holding-dollar"></i></abbr>',
            _("Als bezahlt markiert?"),
        ),
        ordering="is_paid",
        boolean=True,
    )
    def display_is_paid(self):
        return self.is_paid

    def is_correct_payment(self, amount: float, date: datetime):
        """Check if a payment made on a certain date has the correct amount for this order"""

        if amount == self.cached_sum:
            return True

        for condition in self.get_payment_conditions_data():
            if amount == condition["price"] and date <= condition["date"]:
                return True

        return False

    @admin.display(description=_("Bestellung"))
    def __str__(self):
        return self.name()

    def create_email_invoice(self, lang=None):
        # Set invoice date if not present
        if self.invoice_date is None:
            self.invoice_date = timezone.now()
            self.save()

        lang = lang or self.language
        with Language(lang):
            context = {
                "tracking_link": str(self.tracking_link()),
                "trackingdata": bool(self.tracking_number and self.is_shipped),
                "id": str(self.pk),
                "woocommerceid": str(self.woocommerceid),
                "woocommercedata": bool(self.woocommerceid),
            }

            if self.woocommerceid:
                ctx = {
                    "id": str(self.pk),
                    "onlineid": str(self.woocommerceid),
                }
                subject = (
                    gettext("Ihre Bestellung Nr. %(id)s (Online #%(onlineid)s)") % ctx
                )
                filename = (
                    gettext("Rechnung Nr. %(id)s (Online #%(onlineid)s).pdf") % ctx
                )
            else:
                ctx = {
                    "id": str(self.pk),
                }
                subject = gettext("Ihre Bestellung Nr. %(id)s") % ctx
                filename = gettext("Rechnung Nr. %(id)s.pdf") % ctx

            self.email_link_invoice = EMail.objects.create(
                subject=subject,
                to=self.addr_billing_email,
                language=lang,
                html_template="order_invoice.html",
                html_context=context,
                notes=gettext(
                    "Diese E-Mail wurde automatisch aus Bestellung #%d generiert."
                )
                % self.pk,
            )

            self.email_link_invoice.add_attachments(
                Attachment.objects.create_from_binary(
                    filename=filename,
                    content=PDFOrder(self, gettext("Rechnung")).get_pdf(),
                )
            )

            self.save()
            return self.email_link_invoice

    def create_email_shipped(self, lang=None):
        lang = lang or self.language
        with Language(lang):
            context = {
                "tracking_link": str(self.tracking_link()),
                "trackingdata": bool(self.tracking_link() and self.is_shipped),
                "id": str(self.pk),
                "woocommerceid": str(self.woocommerceid),
                "woocommercedata": bool(self.woocommerceid),
            }

            if self.woocommerceid:
                ctx = {
                    "id": str(self.pk),
                    "onlineid": str(self.woocommerceid),
                }
                subject = (
                    gettext("Ihre Lieferung Nr. %(id)s (Online #%(onlineid)s)") % ctx
                )
                filename = (
                    gettext("Lieferschein Nr. %(id)s (Online #%(onlineid)s).pdf") % ctx
                )
            else:
                ctx = {
                    "id": str(self.pk),
                }
                subject = gettext("Ihre Lieferung Nr. %(id)s") % ctx
                filename = gettext("Lieferschein Nr. %(id)s.pdf") % ctx

            self.email_link_shipped = EMail.objects.create(
                subject=subject,
                to=self.addr_shipping_email,
                language=lang,
                html_template="order_supply.html",
                html_context=context,
                notes=gettext(
                    "Diese E-Mail wurde automatisch aus Bestellung #%d generiert."
                )
                % self.pk,
            )

            self.email_link_shipped.add_attachments(
                Attachment.objects.create_from_binary(
                    filename=filename,
                    content=PDFOrder(
                        self, _("Lieferschein"), is_delivery_note=True
                    ).get_pdf(),
                )
            )

            self.save()
            return self.email_link_shipped

    def get_stock_data(self):
        """Get the stock data of all products in this order"""

        return [p.get_stock_data() for p in self.products.all()]

    def email_stock_warning(self):
        email_receiver = settings.get_db_setting("email-stock-warning-receiver")

        if email_receiver:
            warnings = []
            stock = self.get_stock_data()
            for data in stock:
                if data["stock_in_danger"]:
                    warnings.append(data)

            if warnings != []:
                email = EMail.objects.create(
                    subject=gettext("[KMUHelper] - Lagerbestand knapp!"),
                    to=email_receiver,
                    html_template="order_stock_warning.html",
                    html_context={
                        "warnings": warnings,
                    },
                    notes=gettext(
                        "Diese E-Mail wurde automatisch aus Bestellung #%d generiert."
                    )
                    % self.pk,
                )

                success = email.send(
                    headers={"Bestellungs-ID": str(self.pk)},
                )
                return bool(success)
        else:
            log("No email receiver for stock warning set.")
        return None

    class Meta(WooCommerceModelMixin.Meta):
        verbose_name = _("Bestellung")
        verbose_name_plural = _("Bestellungen")

    def duplicate(self):
        new = Order.objects.create(
            customer=self.customer,
            payment_receiver=self.payment_receiver,
            contact_person=self.contact_person,
            pdf_title=self.pdf_title,
            pdf_text=self.pdf_text,
            payment_conditions=self.payment_conditions,
            customer_note=gettext("Kopie aus Bestellung #%d") % self.pk
            + "\n"
            + "-" * 32
            + "\n"
            + self.customer_note,
        )

        for field in constants.ADDR_SHIPPING_FIELDS + constants.ADDR_BILLING_FIELDS:
            setattr(new, field, getattr(self, field))

        for bp in self.products.through.objects.filter(order=self):
            bp.copyto(new)
        for bk in self.fees.through.objects.filter(order=self):
            bk.copyto(new)
        new.save()
        return new

    def copy_to_supply(self):
        new = Supply.objects.create(
            name=gettext("Rücksendung von Bestellung #%d") % self.pk,
        )
        for lp in self.products.through.objects.filter(order=self):
            if lp.linked_product is not None:
                new.products.add(
                    lp.linked_product, through_defaults={"quantity": lp.quantity}
                )
        new.save()
        return new

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-clipboard-list"

    DICT_EXCLUDE_FIELDS = [
        "products",
        "fees",
        "email_link_invoice",
        "email_link_shipped",
        "customer",
        "contact_person",
        "payment_receiver",
        "is_removed_from_stock",
        "is_shipped",
        "is_paid",
        "payment_method",
        "order_key",
    ]


class Fee(CustomModel):
    """Model representing additional costs"""

    PKFILL_WIDTH = 3

    name = models.CharField(
        verbose_name=_("Bezeichnung"),
        max_length=500,
        default=_("Zusätzliche Kosten"),
        help_text=I18N_HELP_TEXT,
    )
    price = models.FloatField(
        verbose_name=_("Preis (exkl. MwSt)"),
        default=0.0,
    )
    vat_rate = models.FloatField(
        verbose_name=_("MwSt-Satz"),
        choices=constants.VAT_RATES,
        default=constants.VAT_RATE_DEFAULT,
    )

    @admin.display(description=_("Bezeichnung"), ordering="name")
    def clean_name(self):
        return langselect(self.name)

    @admin.display(description=_("Kosten"))
    def __str__(self):
        vat_addition = f" + {self.vat_rate}% MwSt" if self.vat_rate else ""
        return f"[{self.pk}] { self.clean_name() } ({ self.price } CHF{vat_addition})"

    class Meta:
        verbose_name = _("Kosten")
        verbose_name_plural = _("Kosten")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-coins"


class Customer(CustomModel, AddressModelMixin, WooCommerceModelMixin):
    """Model representing a customer"""

    PKFILL_WIDTH = 8
    WOOCOMMERCE_URL_FORMAT = "{}/wp-admin/user-edit.php?user_id={}"
    NOTE_RELATION = "customer"

    email = models.EmailField(
        verbose_name=_("E-Mail Adresse"),
        blank=True,
    )
    first_name = models.CharField(
        verbose_name=_("Vorname"),
        max_length=250,
        default="",
        blank=True,
    )
    last_name = models.CharField(
        verbose_name=_("Nachname"),
        max_length=250,
        default="",
        blank=True,
    )
    company = models.CharField(
        verbose_name=_("Firma"),
        max_length=250,
        default="",
        blank=True,
    )
    username = models.CharField(
        verbose_name=_("Benutzername"),
        max_length=50,
        default="",
        blank=True,
    )
    avatar_url = models.URLField(
        verbose_name=_("Avatar URL"),
        blank=True,
        editable=False,
    )
    language = models.CharField(
        verbose_name=_("Sprache"),
        default="de",
        choices=constants.LANGUAGES,
        max_length=2,
    )

    combine_with = models.ForeignKey(
        to="self",
        verbose_name=_("Zusammenfügen mit"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        help_text=_(
            "Dies kann nicht widerrufen werden! Werte im aktuellen Kunden werden bevorzugt."
        ),
    )
    website = models.URLField(
        verbose_name=_("Webseite"),
        default="",
        blank=True,
    )
    note = models.TextField(
        verbose_name=_("Bemerkung"),
        default="",
        blank=True,
    )

    email_link_registered = models.ForeignKey(
        to="EMail",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    @admin.display(description=_("Avatar"), ordering="avatar_url")
    def avatar(self):
        if self.avatar_url:
            return mark_safe('<img src="' + self.avatar_url + '" width="50px">')
        return ""

    @admin.display(description=_("Kunde"))
    def __str__(self):
        s = f"[{self.pkfill()}] "
        if self.woocommerceid:
            s += f"(WC#{self.woocommerceid}) "
        if self.first_name or self.addr_billing_first_name:
            s += f"{self.first_name or self.addr_billing_first_name} "
        if self.last_name or self.addr_billing_last_name:
            s += f"{self.last_name or self.addr_billing_last_name} "
        if self.company or self.addr_billing_company:
            s += f"{self.company or self.addr_billing_company} "
        if self.addr_billing_postcode and self.addr_billing_city:
            s += f"({self.addr_billing_postcode} {self.addr_billing_city})"
        return s

    class Meta(WooCommerceModelMixin.Meta):
        verbose_name = _("Kunde")
        verbose_name_plural = _("Kunden")

    def save(self, **kwargs):
        if self.combine_with:
            self.combine_with_customer(self.combine_with)
            self.combine_with = None
        super().save(**kwargs)

    def combine_with_customer(self, other: "Customer"):
        """Combine this customer with another customer.

        Values in the current customer are preferred. The other customer gets deleted.
        """

        for FIELD_NAME in (
            constants.ADDR_BILLING_FIELDS
            + constants.ADDR_SHIPPING_FIELDS
            + [
                "woocommerceid",
                "email",
                "first_name",
                "last_name",
                "company",
                "username",
                "avatar_url",
                "language",
                "website",
            ]
        ):
            setattr(
                self,
                FIELD_NAME,
                getattr(self, FIELD_NAME) or getattr(other, FIELD_NAME),
            )

        self.note = self.note + "\n" + other.note

        for order in other.orders.all():
            order.customer = self
            order.save()

        if getattr(other, "linked_note", False):
            other_linked_note = other.linked_note
            other_linked_note.linked_customer = (
                None if getattr(self, "linked_note", False) else self
            )
            other_linked_note.save()

        other.delete()

    def create_email_registered(self, lang=None):
        lang = lang or self.language
        with Language(lang):
            context = {
                "customer": {
                    "id": self.pk,
                    "first_name": self.first_name,
                    "last_name": self.last_name,
                    "company": self.company,
                    "email": self.email,
                }
            }

            self.email_link_registered = EMail.objects.create(
                subject=gettext("Registrierung erfolgreich!"),
                to=self.email,
                language=lang,
                html_template="customer_registered.html",
                html_context=context,
                notes=gettext("Diese E-Mail wurde automatisch aus Kunde #%d generiert.")
                % self.pk,
            )

            self.save()
            return self.email_link_registered

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-users"

    DICT_EXCLUDE_FIELDS = ["email_link_registered", "combine_with"]


class Supplier(CustomModel):
    """Model representing a supplier (used only for categorizing)"""

    PKFILL_WIDTH = 4

    abbreviation = models.CharField(
        verbose_name=_("Kürzel"),
        max_length=5,
    )
    name = models.CharField(
        verbose_name=_("Name"),
        max_length=50,
    )

    website = models.URLField(
        verbose_name=_("Webseite"),
        blank=True,
    )
    phone = models.CharField(
        verbose_name=_("Telefon"),
        max_length=50,
        default="",
        blank=True,
    )
    email = models.EmailField(
        verbose_name=_("E-Mail"),
        null=True,
        blank=True,
    )

    address = models.TextField(
        verbose_name=_("Adresse"),
        default="",
        blank=True,
    )
    note = models.TextField(
        verbose_name=_("Notiz"),
        default="",
        blank=True,
    )

    contact_person_name = models.CharField(
        verbose_name=_("Name"),
        max_length=250,
        default="",
        blank=True,
    )
    contact_person_phone = models.CharField(
        verbose_name=_("Telefon"),
        max_length=50,
        default="",
        blank=True,
    )
    contact_person_email = models.EmailField(
        verbose_name=_("E-Mail"),
        null=True,
        default="",
        blank=True,
    )

    @admin.display(description=_("Lieferant"))
    def __str__(self):
        return f"[{self.pk}] {self.name}"

    def assign(self):
        products = Product.objects.filter(supplier=None)
        for product in products:
            product.supplier = self
            product.save()
        return products.count()

    class Meta:
        verbose_name = _("Lieferant")
        verbose_name_plural = _("Lieferanten")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-truck"


class SupplyItem(CustomModel):
    """Model representing the connection between 'Supply' and 'Product'"""

    supply = models.ForeignKey(
        to="Supply",
        on_delete=models.CASCADE,
    )
    product = models.ForeignKey(
        to="Product",
        verbose_name=_("Produkt"),
        on_delete=models.PROTECT,
    )
    quantity = models.IntegerField(
        verbose_name=_("Menge"),
        default=1,
    )

    @admin.display(description=_("Lieferungsposten"))
    def __str__(self):
        return f"[{self.pk}] {self.quantity}x {self.product}"

    class Meta:
        verbose_name = _("Lieferungsposten")
        verbose_name_plural = _("Lieferungsposten")

    objects = models.Manager()


class Supply(CustomModel):
    """Model representing an *incoming* delivery"""

    NOTE_RELATION = "delivery"

    name = models.CharField(
        verbose_name=_("Name"),
        max_length=50,
        default=default_delivery_title,
    )
    date = models.DateField(
        verbose_name=_("Erfasst am"),
        auto_now_add=True,
    )

    supplier = models.ForeignKey(
        to="Supplier",
        verbose_name=_("Lieferant"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    products = models.ManyToManyField(
        to="Product",
        through="SupplyItem",
        through_fields=("supply", "product"),
    )

    is_added_to_stock = models.BooleanField(
        verbose_name=_("Eingelagert?"),
        default=False,
    )

    @admin.display(description=_("Anzahl Produkte"))
    def total_quantity(self):
        return self.products.through.objects.filter(supply=self).aggregate(
            models.Sum("quantity")
        )["quantity__sum"]

    def add_to_stock(self):
        if not self.is_added_to_stock:
            for i in self.products.through.objects.filter(supply=self):
                i.product.stock_current += i.quantity
                i.product.save()
            self.is_added_to_stock = True
            self.save()
            return True
        return False

    @admin.display(description=_("Lieferung"))
    def __str__(self):
        return f"[{self.pk}] {self.name}"

    class Meta:
        verbose_name = _("Lieferung")
        verbose_name_plural = _("Lieferungen")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-truck-ramp-box"


class Note(CustomModel):
    """Model representing a note"""

    name = models.CharField(
        verbose_name=_("Name"),
        max_length=50,
    )
    description = models.TextField(
        verbose_name=_("Beschrieb"),
        default="",
        blank=True,
    )

    done = models.BooleanField(
        verbose_name=_("Erledigt?"),
        default=False,
    )

    priority = models.IntegerField(
        verbose_name=_("Priorität"),
        default=0,
        blank=True,
    )
    created_at = models.DateTimeField(
        verbose_name=_("Erstellt am"),
        auto_now_add=True,
    )

    linked_order = models.OneToOneField(
        to="Order",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="linked_note",
    )
    linked_product = models.OneToOneField(
        to="Product",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="linked_note",
    )
    linked_customer = models.OneToOneField(
        to="Customer",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="linked_note",
    )
    linked_supply = models.OneToOneField(
        to="Supply",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="linked_note",
    )

    @admin.display(description=_("🔗 Notiz"))
    def __str__(self):
        return f"[{self.pk}] {self.name}"

    def links(self):
        text = ""
        if self.linked_order:
            url = reverse(
                "admin:kmuhelper_order_change",
                kwargs={"object_id": self.linked_order.pk},
            )
            text += _("Bestellung")
            text += f" <a href='{url}'>#{self.linked_order.pk}</a><br>"
        if self.linked_product:
            url = reverse(
                "admin:kmuhelper_product_change",
                kwargs={"object_id": self.linked_product.pk},
            )
            text += _("Produkt")
            text += f" <a href='{url}'>#{self.linked_product.pk}</a><br>"
        if self.linked_customer:
            url = reverse(
                "admin:kmuhelper_customer_change",
                kwargs={"object_id": self.linked_customer.pk},
            )
            text += _("Kunde")
            text += f" <a href='{url}'>#{self.linked_customer.pk}</a><br>"
        if self.linked_supply:
            url = reverse(
                "admin:kmuhelper_supply_change",
                kwargs={"object_id": self.linked_supply.pk},
            )
            text += _("Lieferung")
            text += f" <a href='{url}'>#{self.linked_supply.pk}</a><br>"
        return mark_safe(text) or gettext("Diese Notiz hat keine Verknüpfungen.")

    class Meta:
        verbose_name = _("Notiz")
        verbose_name_plural = _("Notizen")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-note-sticky"


class Product(CustomModel, WooCommerceModelMixin):
    """Model representing a product"""

    NOTE_RELATION = "product"

    article_number = models.CharField(
        verbose_name=_("Artikelnummer"),
        max_length=25,
    )

    name = models.CharField(
        verbose_name=_("Name"),
        max_length=500,
        help_text=I18N_HELP_TEXT,
    )
    short_description = models.TextField(
        verbose_name=_("Kurzbeschrieb"),
        default="",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )
    description = models.TextField(
        verbose_name=_("Beschrieb"),
        default="",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )

    quantity_description = models.CharField(
        verbose_name=_("Mengenbezeichnung"),
        max_length=100,
        default="Stück",
        blank=True,
        help_text=I18N_HELP_TEXT,
    )
    selling_price = models.FloatField(
        verbose_name=_("Normalpreis in CHF (exkl. MwSt)"),
        default=0,
    )
    vat_rate = models.FloatField(
        verbose_name=_("MwSt-Satz"),
        choices=constants.VAT_RATES,
        default=constants.VAT_RATE_DEFAULT,
    )

    stock_current = models.IntegerField(
        verbose_name=_("Lagerbestand"),
        default=0,
    )
    stock_target = models.IntegerField(
        verbose_name=_("Soll-Lagerbestand"),
        default=1,
    )

    note = models.TextField(
        verbose_name=_("Bemerkung"),
        default="",
        blank=True,
        help_text=_("Wird nicht gedruckt oder angezeigt; nur für eigene Zwecke."),
    )

    sale_from = models.DateTimeField(
        verbose_name=_("In Aktion von"),
        blank=True,
        null=True,
    )
    sale_to = models.DateTimeField(
        verbose_name=_("In Aktion bis"),
        blank=True,
        null=True,
    )
    sale_price = models.FloatField(
        verbose_name=_("Aktionspreis in CHF (exkl. MwSt)"),
        blank=True,
        null=True,
    )

    datasheet_url = models.CharField(
        verbose_name=_("Datenblattlink"),
        max_length=500,
        default="",
        blank=True,
    )
    image_url = models.URLField(
        verbose_name=_("Bildlink"),
        default="",
        blank=True,
    )

    supplier = models.ForeignKey(
        to="Supplier",
        verbose_name=_("Lieferant"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    supplier_price = models.CharField(
        verbose_name=_("Einkaufspreis"),
        default="",
        blank=True,
        max_length=20,
    )
    supplier_article_number = models.CharField(
        verbose_name=_("Lieferantenartikelnummer"),
        default="",
        blank=True,
        max_length=25,
    )
    supplier_url = models.URLField(
        verbose_name=_("Lieferantenurl (Für Nachbestellungen)"),
        default="",
        blank=True,
    )

    parent = models.ForeignKey(
        to="self",
        related_name="children",
        verbose_name=_("Übergeordnetes Produkt"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    categories = models.ManyToManyField(
        to="ProductCategory",
        through="ProductProductCategoryConnection",
        through_fields=("product", "category"),
        verbose_name=_("Kategorien"),
        related_name="products",
    )

    @admin.display(description=_("Bezeichnung"), ordering="name")
    def clean_name(self, lang="de"):
        return langselect(self.name, lang)

    @admin.display(description=_("Kurzbeschrieb"), ordering="short_description")
    def clean_short_description(self, lang="de"):
        return langselect(self.short_description, lang)

    @admin.display(description=_("Beschrieb"), ordering="description")
    def clean_description(self, lang="de"):
        return langselect(self.description, lang)

    @admin.display(description=_("Aktion?"), boolean=True)
    def is_on_sale(self):
        dt = timezone.now()
        if (not self.sale_price) or (self.selling_price == self.sale_price):
            # No sale price or sale price = regular price
            return None
        if self.sale_from and dt < self.sale_from:
            # Not yet started
            return False
        if self.sale_to and self.sale_to < dt:
            # Already ended
            return False
        return True

    def get_current_price(self):
        return self.sale_price if self.is_on_sale() else self.selling_price

    @admin.display(description=_("Aktueller Preis"))
    def display_current_price(self):
        return formatprice(self.get_current_price()) + " CHF"

    @admin.display(description=_("Nr."), ordering="article_number")
    def display_article_number(self):
        return self.article_number

    @admin.display(description=_("Produktbild"), ordering="image_url")
    def html_image(self):
        if self.image_url:
            return format_html('<img src="{}" width="100px">', self.image_url)
        return ""

    # Stock

    def get_reserved_stock(self):
        return (
            OrderItem.objects.filter(
                order__is_shipped=False, linked_product_id=self.pk
            ).aggregate(models.Sum("quantity"))["quantity__sum"]
            or 0
        )

    def get_incoming_stock(self):
        return (
            SupplyItem.objects.filter(
                supply__is_added_to_stock=False, product__id=self.pk
            ).aggregate(models.Sum("quantity"))["quantity__sum"]
            or 0
        )

    def get_stock_data(self, includemessage=False):
        """Get the stock and product information as a dictionary"""

        p_id = self.pk
        p_name = self.clean_name()
        p_article_number = self.article_number

        n_current = self.stock_current
        n_going = self.get_reserved_stock()
        n_coming = self.get_incoming_stock()
        n_min = self.stock_target

        data = {
            "product": {
                "id": p_id,
                "article_number": p_article_number,
                "name": p_name,
            },
            "stock": {
                "current": n_current,
                "going": n_going,
                "coming": n_coming,
                "min": n_min,
            },
            "stock_overbooked": n_current - n_going < 0,
            "stock_in_danger": n_current - n_going < n_min,
        }

        if includemessage:
            t_current = gettext("Aktuell")
            t_going = gettext("Ausgehend")
            t_coming = gettext("Eingehend")

            stockstring = f"{t_current}: { n_current } | {t_going}: { n_going }"
            if t_coming:
                stockstring += f" | {t_coming}: { n_coming }"

            adminurl = reverse(f"admin:kmuhelper_product_change", args=[self.pk])
            adminlink = format_html('<a href="{}">{}</a>', adminurl, p_name)

            formatdata = (adminlink, p_article_number, stockstring)

            if data["stock_overbooked"]:
                msg = gettext("Zu wenig Lagerbestand bei")
                data["message"] = format_html('{} "{}" [{}]: {}', msg, *formatdata)
            elif data["stock_in_danger"]:
                msg = gettext("Knapper Lagerbestand bei")
                data["message"] = format_html('{} "{}" [{}]: {}', msg, *formatdata)

        return data

    def show_stock_warning(self, request):
        data = self.get_stock_data(includemessage=True)

        if data["stock_overbooked"]:
            messages.error(request, data["message"])
        elif data["stock_in_danger"]:
            messages.warning(request, data["message"])

    @admin.display(description=_("Produkt"))
    def __str__(self):
        return f"[{self.pk}] {self.article_number} - {self.clean_name()}"

    class Meta(WooCommerceModelMixin.Meta):
        verbose_name = _("Produkt")
        verbose_name_plural = _("Produkte")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-cubes"


class ProductCategory(CustomModel, WooCommerceModelMixin):
    """Model representing a category for products"""

    WOOCOMMERCE_URL_FORMAT = (
        "{}/wp-admin/term.php?taxonomy=product_cat&tag_ID={}&post_type=product"
    )
    PKFILL_WIDTH = 4

    name = models.CharField(
        verbose_name=_("Bezeichnung"),
        max_length=250,
        default="",
    )
    description = models.TextField(
        verbose_name=_("Beschrieb"),
        default="",
        blank=True,
    )
    image_url = models.URLField(
        verbose_name=_("Bildlink"),
        blank=True,
    )

    parent_category = models.ForeignKey(
        to="self",
        verbose_name=_("Übergeordnete Kategorie"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    @admin.display(description=_("Bild"), ordering="image_url")
    def html_image(self):
        if self.image_url:
            return format_html('<img src="{}" width="100px">', self.image_url)
        return ""

    @admin.display(description=_("Anzahl Produkte"))
    def total_quantity(self):
        return self.products.count()

    @admin.display(description=_("Bezeichnung"), ordering="name")
    def clean_name(self):
        return langselect(self.name)

    @admin.display(description=_("Beschrieb"), ordering="description")
    def clean_description(self):
        return langselect(self.description)

    @admin.display(description=_("Kategorie"))
    def __str__(self):
        return f"[{self.pk}] {self.clean_name()}"

    class Meta(WooCommerceModelMixin.Meta):
        verbose_name = _("Produktkategorie")
        verbose_name_plural = _("Produktkategorien")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-folder-tree"


class ProductProductCategoryConnection(CustomModel):
    """Model representing the connection between 'Product' and 'ProductCategory'"""

    product = models.ForeignKey(
        to="Product",
        verbose_name=_("Produkt"),
        on_delete=models.CASCADE,
    )
    category = models.ForeignKey(
        to="ProductCategory",
        verbose_name=_("Produktkategorie"),
        on_delete=models.CASCADE,
    )

    @admin.display(description=_("Produkt-Kategorie-Verknüpfung"))
    def __str__(self):
        return f"{self.category.clean_name()} <- [{self.pk}] -> {self.product}"

    class Meta:
        verbose_name = _("Produkt-Kategorie-Verknüpfung")
        verbose_name_plural = _("Produkt-Kategorie-Verknüpfungen")

    objects = models.Manager()


class PaymentReceiver(CustomModel):
    """Model representing a payment receiver for the qr bill"""

    PKFILL_WIDTH = 3

    # Internal

    internal_name = models.CharField(
        verbose_name=_("Interne Bezeichnung"),
        max_length=250,
        default="",
    )

    # Payment

    mode = models.CharField(
        verbose_name=_("Modus"),
        max_length=15,
        choices=[
            ("QRR", _("QR-Referenz")),
            ("NON", _("Ohne Referenz")),
        ],
        default="QRR",
    )

    qriban = models.CharField(
        verbose_name=_("QR-IBAN"),
        max_length=21 + 5,
        validators=[
            RegexValidator(
                r"^(CH|LI)[0-9]{2}\s3[0-9]{3}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{1}$",
                _("Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x"),
            ),
        ],
        help_text=_("QR-IBAN mit Leerzeichen (Nur verwendet im Modus 'QR-Referenz')"),
        blank=True,
        default="",
    )
    iban = models.CharField(
        verbose_name=_("IBAN"),
        max_length=21 + 5,
        validators=[
            RegexValidator(
                r"^(CH|LI)[0-9]{2}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{1}$",
                _("Bite benutze folgendes Format: (CH|LI)pp 3xxx xxxx xxxx xxxx x"),
            ),
        ],
        help_text=_("IBAN mit Leerzeichen (Nur verwendet im Modus 'Ohne Referenz')"),
        blank=True,
        default="",
    )

    # Display information

    display_name = models.CharField(
        verbose_name=_("Anzeigename"),
        max_length=70,
        default="",
        help_text=_("Wird oben auf der Rechnung angezeigt."),
    )
    display_address_1 = models.CharField(
        verbose_name=_("Adresszeile 1"),
        max_length=70,
        default="",
        help_text=_("Wird oben auf der Rechnung angezeigt."),
    )
    display_address_2 = models.CharField(
        verbose_name=_("Adresszeile 2"),
        max_length=70,
        default="",
        help_text=_("Wird oben auf der Rechnung angezeigt."),
    )

    website = models.URLField(
        verbose_name=_("Webseite"),
        default="",
        blank=True,
        help_text=_("Auf der Rechnung ersichtlich, sofern vorhanden!"),
    )

    logourl = models.URLField(
        verbose_name=_("Logo (URL)"),
        validators=[
            RegexValidator(
                r"""^[0-9a-zA-Z\-\.\|\?\(\)\*\+&"'_:;/]+\.(png|jpg)$""",
                _("""Nur folgende Zeichen gestattet: 0-9a-zA-Z-_.:;/|?&()"'*+ - """)
                + _("""Muss auf .jpg/.png enden."""),
            ),
        ],
        help_text=_("URL eines Bildes (.jpg/.png) - Wird auf die Rechnung gedruckt."),
        blank=True,
        default="",
    )

    swiss_uid = models.CharField(
        verbose_name=_("Firmen-UID"),
        max_length=15,
        validators=[
            RegexValidator(
                r"^CHE-[0-9]{3}\.[0-9]{3}\.[0-9]{3}$",
                _("Bite benutze folgendes Format: CHE-123.456.789"),
            )
        ],
        help_text=_("UID der Firma - Format: CHE-123.456.789 (Mehrwertsteuernummer)"),
        blank=True,
        default="",
    )

    invoice_display_mode = models.CharField(
        verbose_name=_("Anzeigemodus für Rechnungen"),
        max_length=25,
        choices=constants.INVOICE_DISPLAY_MODES,
        default="business_orders",
        help_text=_("Je nach Modus werden andere Daten auf der Rechnung angezeigt."),
    )

    # Payment information

    invoice_name = models.CharField(
        verbose_name=_("Kontoinhaber"),
        max_length=70,
        help_text=_("Name / Firma"),
    )
    invoice_street = models.CharField(
        verbose_name=pgettext_lazy("address", "Strasse"),
        max_length=70,
        help_text=_("Strasse oder Postfach"),
    )
    invoice_street_nr = models.CharField(
        verbose_name=pgettext_lazy("address", "Hausnummer"),
        max_length=16,
        blank=True,
    )
    invoice_postcode = models.CharField(
        verbose_name=pgettext_lazy("address", "Postleitzahl"),
        max_length=16,
    )
    invoice_city = models.CharField(
        verbose_name=pgettext_lazy("address", "Ort"),
        max_length=35,
    )
    invoice_country = models.CharField(
        verbose_name=_("Land"),
        max_length=2,
        choices=constants.COUNTRIES,
        default="CH",
    )

    # Default

    is_default = models.BooleanField(
        verbose_name=_("Standard?"),
        default=False,
        help_text=_(
            "Aktivieren, um diesen Zahlungsempfänger als Standard zu setzen. Der Standard-Zahlungsempfänger wird bei "
            "der Erstellung einer neuen Bestellung sowie beim Import einer Bestellung von WooCommerce verwendet."
        ),
    )

    # Validation

    @classmethod
    def _check_iban(cls, iban: str, qr_required=False):
        try:
            b = ""
            # Translate letters to numbers
            for i in (0, 1):
                a = str(iban)[i].upper()
                if a not in string.ascii_uppercase:
                    return False
                b += str(ord(a) - 55)
            # Select only digits
            num = "".join([z for z in str(iban)[2:] if z in string.digits])
            # Check if QR-IBAN is required
            if qr_required and not num[2] == "3":
                return False
            # Validate IBAN
            result = int(int(num[2:] + b + num[:2]) % 97) == 1
            return result
        except IndexError:
            return False

    @classmethod
    def _check_uid(cls, uid: str):
        try:
            u = uid.split("-")[1].replace(".", "")
            p = 11 - (
                (
                    (int(u[0]) * 5)
                    + (int(u[1]) * 4)
                    + (int(u[2]) * 3)
                    + (int(u[3]) * 2)
                    + (int(u[4]) * 7)
                    + (int(u[5]) * 6)
                    + (int(u[6]) * 5)
                    + (int(u[7]) * 4)
                )
                % 11
            )
            return int(u[8]) == p
        except Exception as e:
            log("Error while validating UID:", e)
            return False

    def has_valid_qr_iban(self):
        return self._check_iban(self.qriban, qr_required=True)

    def has_valid_iban(self):
        return self._check_iban(self.iban)

    def has_valid_uid(self):
        return self._check_uid(self.swiss_uid)

    # Properties

    @property
    @admin.display(description=_("Interne Bezeichnung"))
    def admin_name(self):
        return self.internal_name or self.display_name or self.invoice_name

    @property
    @admin.display(description=_("IBAN"))
    def active_iban(self):
        return self.qriban if self.mode == "QRR" else self.iban

    # More

    @admin.display(description=_("Zahlungsempfänger"))
    def __str__(self):
        return f"[{self.pk}] {self.admin_name}"

    def clean(self):
        super().clean()

        errors = {}

        if self.mode == "QRR" and not self.has_valid_qr_iban():
            errors["qriban"] = ValidationError(
                _("Im Modus 'QR-Referenz' muss eine gültige QR-IBAN angegeben werden!")
            )
        if self.mode == "NON" and not self.has_valid_iban():
            errors["iban"] = ValidationError(
                _("Im Modus 'Ohne Referenz' muss eine gültige IBAN angegeben werden!")
            )
        if self.swiss_uid and not self.has_valid_uid():
            errors["swiss_uid"] = ValidationError(
                _("Die UID ist ungültig!"), code="invalid"
            )
        if self.logourl:
            try:
                response = requests.get(self.logourl)
                response.raise_for_status()
            except requests.RequestException as e:
                errors["logourl"] = ValidationError(
                    _("An dieser Adresse konnte kein Bild abgerufen werden! Fehler: %s")
                    % str(e)
                )

        if errors:
            raise ValidationError(errors)

    class Meta:
        verbose_name = _("Zahlungsempfänger")
        verbose_name_plural = _("Zahlungsempfänger")

    objects = models.Manager()

    ADMIN_ICON = "fa-solid fa-hand-holding-dollar"



# Generated by Django 5.0.4 on 2024-04-06 20:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0111_alter_order_tracking_number"),
    ]

    operations = [
        migrations.AddField(
            model_name="product",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="children",
                to="kmuhelper.product",
                verbose_name="Übergeordnetes Produkt",
            ),
        ),
    ]



import base64
import hmac
import hashlib
import secrets
import requests

from kmuhelper import settings


def is_connected() -> bool:
    """Check if WooCommerce has been connected"""

    return bool(settings.get_secret_db_setting("wc-url", False))


def base64_hmac_sha256(key: bytes, message: bytes) -> bytes:
    """Calculate HMAC-SHA256 hash of message using key"""

    return base64.b64encode(hmac.new(key, message, hashlib.sha256).digest())


def random_secret() -> str:
    """Generate a random secret"""

    return secrets.token_urlsafe(32)


def test_wc_url(url: str) -> bool:
    """Test if a URL is a WooCommerce store"""

    try:
        response = requests.get(url.rstrip("/") + "/wp-json/wc/v3")
        response.raise_for_status()
    except requests.exceptions.RequestException:
        return False

    return True



# Generated by Django 5.0 on 2023-12-19 15:24

from django.db import migrations, models


def set_initial_defaults(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    ContactPerson = apps.get_model("kmuhelper", "ContactPerson")
    PaymentReceiver = apps.get_model("kmuhelper", "PaymentReceiver")
    Order = apps.get_model("kmuhelper", "Order")

    contact_person = None
    payment_receiver = None

    if Order.objects.using(db_alias).exists():
        newest_order = Order.objects.using(db_alias).order_by("-date").first()
        contact_person = newest_order.contact_person
        payment_receiver = newest_order.payment_receiver
    else:
        if ContactPerson.objects.using(db_alias).exists():
            contact_person = ContactPerson.objects.using(db_alias).first()
        if PaymentReceiver.objects.using(db_alias).exists():
            payment_receiver = PaymentReceiver.objects.using(db_alias).first()

    if contact_person:
        contact_person.is_default = True
        contact_person.save()
    if payment_receiver:
        payment_receiver.is_default = True
        payment_receiver.save()


class Migration(migrations.Migration):
    dependencies = [
        ("kmuhelper", "0108_migrate_vat_rates_for_2024"),
    ]

    operations = [
        migrations.AddField(
            model_name="contactperson",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="Aktivieren, um diese Kontaktperson als Standard zu setzen. Die Standard-Kontaktperson wird bei der Erstellung einer neuen Bestellung sowie beim Import einer Bestellung von WooCommerce verwendet.",
                verbose_name="Standard?",
            ),
        ),
        migrations.AddField(
            model_name="paymentreceiver",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="Aktivieren, um diesen Zahlungsempfänger als Standard zu setzen. Der Standard-Zahlungsempfänger wird bei der Erstellung einer neuen Bestellung sowie beim Import einer Bestellung von WooCommerce verwendet.",
                verbose_name="Standard?",
            ),
        ),
        migrations.RunPython(
            set_initial_defaults, migrations.RunPython.noop, elidable=True
        ),
    ]
